"""Entropy Pooling estimator."""

# Copyright (c) 2025
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# Credits: Vincent Maladière, Matteo Manzi, Carlo Nicolini
# SPDX-License-Identifier: BSD-3-Clause

import operator
import re
import warnings
from collections.abc import Sequence
from typing import TYPE_CHECKING

import cvxpy as cp
import numpy as np
import numpy.typing as npt
import scipy.optimize as sco
import scipy.sparse.linalg as scl
import scipy.special as scs
import scipy.stats as sts
import sklearn.utils.metadata_routing as skm
import sklearn.utils.validation as skv

import skfolio.measures as sm
import skfolio.typing as skt
from skfolio.exceptions import SolverError
from skfolio.measures import (
    ExtraRiskMeasure,
    PerfMeasure,
    RiskMeasure,
)
from skfolio.prior._base import BasePrior, ReturnDistribution
from skfolio.prior._empirical import EmpiricalPrior
from skfolio.utils.equations import equations_to_matrix
from skfolio.utils.tools import check_estimator, default_asset_names, input_to_array


class EntropyPooling(BasePrior):
    r"""Entropy Pooling estimator.

    Entropy Pooling, introduced by Attilio Meucci in 2008 as a generalization of the
    Black-Litterman framework, is a nonparametric method for adjusting a baseline
    ("prior") probability distribution to incorporate user-defined views by finding the
    posterior distribution closest to the prior while satisfying those views.

    User-defined views can be **elicited** from domain experts or **derived** from
    quantitative analyses.

    Grounded in information theory, it updates the distribution in the least-informative
    way by minimizing the Kullback-Leibler divergence (relative entropy) under the
    specified view constraints.

    Mathematically, the problem is formulated as:

    .. math::

       \begin{aligned}
       \min_{\mathbf{q}} \quad & \sum_{i=1}^T q_i \log\left(\frac{q_i}{p_i}\right) \\
       \text{subject to} \quad & \sum_{i=1}^T q_i = 1 \quad \text{(normalization constraint)} \\
                               & \mathbb{E}_q[f_j(X)] = v_j \quad(\text{or } \le v_j, \text{ or } \ge v_j), \quad j = 1,\dots,k, \text{(view constraints)} \\
                               & q_i \ge 0, \quad i = 1, \dots, T
       \end{aligned}

    Where:

    - :math:`T` is the number of observations (number of scenarios).
    - :math:`p_i` is the prior probability of scenario :math:`x_i`.
    - :math:`q_i` is the posterior probability of scenario :math:`x_i`.
    - :math:`X` is the scenario matrix of shape (n_observations, n_assets).
    - :math:`f_j` is the j :sup:`th` view function.
    - :math:`v_j` is the target value imposed by the j :sup:`th` view.
    - :math:`k` is the total number of views.

    The `skfolio` implementation supports the following views:
        * Equalities
        * Inequalities
        * Ranking
        * Linear combinations (e.g. relative views)
        * Views on groups of assets

    On the following measures:
        * Mean
        * Variance
        * Skew
        * Kurtosis
        * Correlation
        * Value-at-Risk (VaR)
        * Conditional Value-at-Risk (CVaR)

    Notes
    -----
    Entropy Pooling re-weights the sample probabilities of the prior distribution and is
    therefore constrained by the support (completeness) of that distribution. For
    example, if the historical distribution contains no returns below -10% for a given
    asset, we cannot impose a CVaR view of 15%: no matter how we adjust the sample
    probabilities, such tail data do not exist.

    Therefore, to impose extreme views on a sparse historical distribution, one must
    generate synthetic data. In that case, the EP posterior is only as reliable as the
    synthetic scenarios. It is thus essential to use a generator capable of
    extrapolating tail dependencies, such as :class:`~skfolio.distribution.VineCopula`,
    to model joint extreme events accurately.

    Two methods are available:
        * Dual form: solves the Fenchel dual of the EP problem using Truncated
          Newton Constrained method.
        * Primal form: solves the original relative-entropy projection in
          probability-space via interior-point algorithms.

    See the solver parameter's docstring for full details on available solvers and
    options.

    To handle nonlinear views, constraints are linearized by fixing the relevant asset
    moments (e.g., means or variances) and then solved via **nested entropic tilting**.
    At each stage, the KL-divergence is minimized relative to the original prior,
    while nesting all previously enforced (linearized) views into the feasible set:

    * Stage 1: impose views on asset means, VaR and CVaR.
    * Stage 2: carry forward Stage 1 constraints and add variance, fixing the
      mean at its Stage 1 value.
    * Stage 3: carry forward Stage 2 constraints and add skewness, kurtosis and
      pairwise correlations, fixing both mean and variance at their Stage 2 values.

    Because each entropic projection nests the prior views, every constraint from
    earlier stages is preserved as new ones are added, yielding a final distribution
    that satisfies all original nonlinear views while staying as close as possible to
    the original prior.

    Only the necessary moments are fixed. Slack variables with an L1 norm penalty are
    introduced to avoid solver infeasibility that may arise from overly tight
    constraints.

    CVaR view constraints cannot be directly expressed as linear functions of the
    posterior probabilities. Therefore, when CVaR views are present, the EP problem is
    solved by recursively solving a series of convex programs that approximate the
    nonlinear CVaR constraint.

    This implementation improves upon Meucci's algorithm [1]_ by formulating the problem
    in continuous space as a function of the dual variables etas (VaR levels), rather
    than searching over discrete tail sizes. This formulation not only handles the
    CVaR constraint more directly but also supports multiple CVaR views on different
    assets.

    Although the overall problem is convex in the dual variables etas, it remains
    non-smooth due to the presence of the positive-part operator in the CVaR
    definition. Consequently, we employ derivative-free optimization methods.
    Specifically, for a single CVaR view we use a one-dimensional root-finding
    method (Brent's method), and for the multivariate case (supporting multiple
    CVaR views) we use Powell's method for derivative-free convex descent.

    Parameters
    ----------
    prior_estimator : BasePrior, optional
        Estimator of the asset's prior distribution, fitted from a
        :ref:`prior estimator <prior>`. The default (`None`) is to use the
        empirical prior :class:`~skfolio.prior.EmpiricalPrior()`. To perform Entropy
        Pooling on synthetic data, you can use :class:`~skfolio.prior.SyntheticData`
        by setting `prior_estimator = SyntheticData()`.

    mean_views : list[str], optional
        Views on asset means.
        The views must match any of following patterns:

            * `"ref1 >= a"`
            * `"ref1 == b"`
            * `"ref1 <= ref1"`
            * `"ref1 >= a * prior(ref1)"`
            * `"ref1 == b * prior(ref2)"`
            * `"a * ref1 + b * ref2 + c <= d * ref3"`

        With `"ref1"`, `"ref2"` ... the assets names or the groups names provided
        in the parameter `groups`. Assets names can be referenced without the need of
        `groups` if the input `X` of the `fit` method is a DataFrame with assets names
        in columns. Otherwise, the default asset names `x0, x1, ...` are assigned.
        By using the term `prior(...)`, you can reference the asset prior mean.

        For example:

            * `"SPX >= 0.0015"` --> The mean of SPX is greater than 0.15% (daily mean if
              `X` is daily)
            * `"SX5E == 0.002"` --> The mean of SX5E equals 0.2%
            * `"AAPL <= 0.003"` --> The mean of AAPL is less than to 0.3%
            * `"SPX <= SX5E"` --> Ranking view: the mean of SPX is less than SX5E
            * `"SPX >= 1.5 * prior(SPX)"` --> The mean of SPX increases by at least 50% (versus its prior)
            * `"SX5E == 2 * prior(SX5E)"` --> The mean of SX5E doubles (versus its prior)
            * `"AAPL <= 0.8 * prior(SPX)"` --> The mean of AAPL is less than 0.8 times the SPX prior
            * `"SX5E + SPX >= 0"` --> The sum of SX5E and SPX mean is greater than zero
            * `"US == 0.007"` --> The sum of means of US assets equals 0.7%
            * `"Equity == 3 * Bond"` --> The sum of means of Equity assets equals
              three times the  sum of means of Bond assets.
            * `"2*SPX + 3*Europe <= Bond + 0.05"` --> Mixing assets and group mean views

    variance_views : list[str], optional
        Views on asset variances.
        It supports the same patterns as `mean_views`.

        For example:

            * `"SPX >= 0.0009"` --> SPX variance is greater than 0.0009 (daily)
            * `"SX5E == 1.5 * prior(SX5E)"` --> SX5E variance increases by 150% (versus
              its prior)

    skew_views : list[str], optional
        Views on asset skews.
        It supports the same patterns as `mean_views`.

        For example:

            * `"SPX >= 2.0"` --> SPX skew is greater than 2.0
            * `"SX5E == 1.5 * prior(SX5E)"` --> SX5E skew increases by 150% (versus its
              prior)

    kurtosis_views : list[str], optional
        Views on asset kurtosis.
        It supports the same patterns as `mean_views`.

        For example:

            * `"SPX >= 9.0"` --> SPX kurtosis is greater than 9.0
            * `"SX5E == 1.5 * prior(SX5E)"` --> SX5E kurtosis increases by 150% (versus
              its prior)

    correlation_views : list[str], optional
        Views on asset correlations.
        The views must match any of following patterns:

            * `"(asset1, asset2) >= a"`
            * `"(asset1, asset2) == a"`
            * `"(asset1, asset2) <= a"`
            * `"(asset1, asset2) >= a * prior(asset1, asset2)"`
            * `"(asset1, asset2) == a * prior(asset1, asset2)"`
            * `"(asset1, asset2) <= a * prior(asset1, asset2)"`

        For example:

            * `"(SPX, SX5E) >= 0.8"` --> the correlation between SPX and SX5E is greater than 80%
            * `"(SPX, SX5E) == 1.5 * prior(SPX, SX5E)"` --> the correlation between SPX
              and SX5E increases by 150% (versus its prior)

    value_at_risk_views : list[str], optional
        Views on asset Value-at-Risks (VaR).

        For example:

            * `"SPX >= 0.03"` --> SPX VaR is greater than 3%
            * `"SX5E == 1.5 * prior(SX5E)"` --> SX5E VaR increases by 150% (versus its prior)

    cvar_views : list[str], optional
        Views on asset Conditional Value-at-Risks (CVaR).
        It only supports equalities.

        For example:

            * `"SPX == 0.05"` --> SPX CVaR equals 5%
            * `"SX5E == 1.5 * prior(SX5E)"` --> SX5E CVaR increases by 150% (versus its prior)

    value_at_risk_beta : float, default=0.95
        Confidence level for VaR views, by default 95%.

    cvar_beta : float, default=0.95
        Confidence level for CVaR views, by default 95%.

    groups : dict[str, list[str]] or array-like of strings of shape (n_groups, n_assets), optional
        Asset grouping for use in group-based views. If a dict is provided, keys are
        asset names and values are lists of group labels; then `X` must be a DataFrame
        whose columns match those asset names.

        For example:

            * `groups = {"SX5E": ["Equity", "Europe"], "SPX": ["Equity", "US"], "TLT": ["Bond", "US"]}`
            * `groups = [["Equity", "Equity", "Bond"], ["Europe", "US", "US"]]`

    solver : str, default="TNC"
        The solver to use.

        - "TNC" (default) solves the entropic-pooling dual via SciPy's Truncated Newton
          Constrained method. By exploiting the smooth Fenchel dual and its
          closed-form gradient, it operates in :math:`\mathbb{R}^k` (the number of
          constraints) rather than :math:`\mathbb{R}^T` (the number of scenarios),
          yielding an order-of-magnitude speedup over primal CVXPY interior-point
          solvers.

        - CVXPY solvers (e.g. "CLARABEL") solve the entropic-pooling problem in its
          primal form using interior-point methods. While they tend to be slower than
          the dual-based approach, they often achieve higher accuracy by enforcing
          stricter primal feasibility and duality-gap tolerances. See the CVXPY
          documentation for supported solvers:
          https://www.cvxpy.org/tutorial/advanced/index.html#choosing-a-solver.

    solver_params : dict, optional
        Additional parameters to pass to the chosen solver.

        - When using **SciPy TNC**, supported options include (but are not limited to)
          `gtol`, `ftol`, `eps`, `maxfun`, `maxCGit`, `stepmx`, `disp`. See the SciPy
          documentation for a full list and descriptions:
          https://docs.scipy.org/doc/scipy/reference/optimize.minimize-tnc.html

        - When using a **CVXPY** solver (e.g. ``"CLARABEL"``), supply any
          solver-specific parameters here. Refer to the CVXPY solver guide for
          details: https://www.cvxpy.org/tutorial/solvers

    Attributes
    ----------
    return_distribution_ : ReturnDistribution
        Fitted :class:`~skfolio.prior.ReturnDistribution` to be used by the optimization
        estimators, containing the assets distribution, moments estimation and the EP
        posterior probabilities (sample weights).

    relative_entropy_ : float
        The KL-divergence between the posterior and prior distributions.

    effective_number_of_scenarios_ : float
        Effective number of scenarios defined as the perplexity of sample weight
        (exponential of entropy).

    prior_estimator_ : BasePrior
        Fitted `prior_estimator`.

    n_features_in_ : int
       Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
       Names of features seen during `fit`. Defined only when `X`
       has feature names that are all strings.

    References
    ----------
    .. [1] "Fully Flexible Extreme Views",
            Journal of Risk, Meucci, Ardia & Keel (2011)

    .. [2] "Fully Flexible Views: Theory and Practice",
            Risk, Meucci (2013).

    .. [3] "Effective Number of Scenarios in Fully Flexible Probabilities",
            GARP Risk Professional, Meucci (2012)

    .. [4] "I-Divergence Geometry of Probability Distributions and Minimization
            Problems", The Annals of Probability, Csiszar (1975)


    Examples
    --------
    For a full tutorial on entropy pooling, see :ref:`sphx_glr_auto_examples_entropy_pooling_plot_1_entropy_pooling.py`.

    >>> from skfolio import RiskMeasure
    >>> from skfolio.datasets import load_sp500_dataset
    >>> from skfolio.preprocessing import prices_to_returns
    >>> from skfolio.prior import EntropyPooling
    >>> from skfolio.optimization import HierarchicalRiskParity
    >>>
    >>> prices = load_sp500_dataset()
    >>> prices = prices[["AMD", "BAC", "GE", "JNJ", "JPM", "LLY", "PG"]]
    >>> X = prices_to_returns(prices)
    >>>
    >>> groups = {
    ...     "AMD": ["Technology", "Growth"],
    ...     "BAC": ["Financials", "Value"],
    ...     "GE": ["Industrials", "Value"],
    ...     "JNJ": ["Healthcare", "Defensive"],
    ...     "JPM": ["Financials", "Income"],
    ...     "LLY": ["Healthcare", "Defensive"],
    ...     "PG": ["Consumer", "Defensive"],
    ... }
    >>>
    >>> entropy_pooling = EntropyPooling(
    ...     mean_views=[
    ...         "JPM == -0.002",
    ...         "PG >= LLY",
    ...         "BAC >= prior(BAC) * 1.2",
    ...         "Financials == 2 * Growth",
    ...     ],
    ...     variance_views=[
    ...         "BAC == prior(BAC) * 4",
    ...     ],
    ...     correlation_views=[
    ...         "(BAC,JPM) == 0.80",
    ...         "(BAC,JNJ) <= prior(BAC,JNJ) * 0.5",
    ...     ],
    ...     skew_views=[
    ...         "BAC == -0.05",
    ...     ],
    ...     cvar_views=[
    ...         "GE == 0.08",
    ...     ],
    ...     cvar_beta=0.90,
    ...     groups=groups,
    ... )
    >>>
    >>> entropy_pooling.fit(X)
    EntropyPooling(correlation_views=...
    >>>
    >>> print(entropy_pooling.relative_entropy_)
    0.18...
    >>> print(entropy_pooling.effective_number_of_scenarios_)
    6876.67...
    >>> print(entropy_pooling.return_distribution_.sample_weight)
    [0.000103...  0.000093... ... 0.000103...  0.000108...]
    >>>
    >>> # CVaR Hierarchical Risk Parity optimization on Entropy Pooling
    >>> model = HierarchicalRiskParity(
    ...     risk_measure=RiskMeasure.CVAR,
    ...     prior_estimator=entropy_pooling
    ... )
    >>> model.fit(X)
    HierarchicalRiskParity(prior_estimator=...
    >>> print(model.weights_)
    [0.073... 0.0541... ... 0.200...]
    >>>
    >>> # Stress Test the Portfolio
    >>> entropy_pooling = EntropyPooling(cvar_views=["AMD == 0.10"])
    >>> entropy_pooling.fit(X)
    EntropyPooling(cvar_views=['AMD == 0.10'])
    >>>
    >>> stressed_dist = entropy_pooling.return_distribution_
    >>>
    >>> stressed_ptf = model.predict(stressed_dist)
    """

    relative_entropy_: float
    effective_number_of_scenarios_: float
    prior_estimator_: BasePrior
    n_features_in_: int
    feature_names_in_: np.ndarray

    if TYPE_CHECKING:
        _returns: np.ndarray
        _prior_sample_weight: np.ndarray
        _groups: np.ndarray
        _is_fixed_mean: np.ndarray
        _is_fixed_variance: np.ndarray
        _constraints: dict[str, list[np.ndarray] | None]

    def __init__(
        self,
        prior_estimator: BasePrior | None = None,
        mean_views: list[str] | None = None,
        variance_views: list[str] | None = None,
        correlation_views: list[str] | None = None,
        skew_views: list[str] | None = None,
        kurtosis_views: list[str] | None = None,
        value_at_risk_views: list[str] | None = None,
        cvar_views: list[str] | None = None,
        value_at_risk_beta: float = 0.95,
        cvar_beta: float = 0.95,
        groups: skt.Groups | None = None,
        solver: str = "TNC",
        solver_params: dict | None = None,
    ):
        self.prior_estimator = prior_estimator
        self.mean_views = mean_views
        self.variance_views = variance_views
        self.correlation_views = correlation_views
        self.skew_views = skew_views
        self.kurtosis_views = kurtosis_views
        self.value_at_risk_views = value_at_risk_views
        self.cvar_views = cvar_views
        self.value_at_risk_beta = value_at_risk_beta
        self.cvar_beta = cvar_beta
        self.groups = groups
        self.solver = solver
        self.solver_params = solver_params

    def get_metadata_routing(self):
        router = skm.MetadataRouter(owner=self.__class__.__name__).add(
            prior_estimator=self.prior_estimator,
            method_mapping=skm.MethodMapping().add(caller="fit", callee="fit"),
        )
        return router

    def fit(self, X: npt.ArrayLike, y=None, **fit_params) -> "EntropyPooling":
        """Fit the Entropy Pooling estimator.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
           Price returns of the assets.

        y : Ignored
           Not used, present for API consistency by convention.

        **fit_params : dict
           Parameters to pass to the underlying estimators.
           Only available if `enable_metadata_routing=True`, which can be
           set by using ``sklearn.set_config(enable_metadata_routing=True)``.
           See :ref:`Metadata Routing User Guide <metadata_routing>` for
           more details.

        Returns
        -------
        self : EntropyPooling
           Fitted estimator.
        """
        routed_params = skm.process_routing(self, "fit", **fit_params)

        # Validation
        skv.validate_data(self, X)

        self.prior_estimator_ = check_estimator(
            self.prior_estimator,
            default=EmpiricalPrior(),
            check_type=BasePrior,
        )
        # Fitting prior estimator
        self.prior_estimator_.fit(X, y, **routed_params.prior_estimator.fit)
        # Prior distribution
        self._returns = self.prior_estimator_.return_distribution_.returns
        n_observations, n_assets = self._returns.shape
        self._prior_sample_weight = (
            self.prior_estimator_.return_distribution_.sample_weight
        )
        if self._prior_sample_weight is None:
            self._prior_sample_weight = np.ones(n_observations) / n_observations

        assets_names = getattr(self, "feature_names_in_", default_asset_names(n_assets))
        if self.groups is None:
            self._groups = np.asarray([assets_names])
        else:
            self._groups = input_to_array(
                items=self.groups,
                n_assets=n_assets,
                fill_value="",
                dim=2,
                assets_names=assets_names,
                name="groups",
            )

        # Init problem variables
        self._is_fixed_mean = np.zeros(n_assets, dtype=bool)
        self._is_fixed_variance = np.zeros(n_assets, dtype=bool)
        self._constraints = {
            "equality": None,
            "inequality": None,
            "fixed_equality": None,
            "cvar_equality": None,
        }

        # Step 1: Mean, VaR and CVaR
        self._add_mean_views()
        self._add_value_at_risk_views()
        sample_weight = self._solve_with_cvar()

        # Step 2:  Mean, VaR, CVaR and Variance
        if self.variance_views is not None:
            # Get mean from Step 1
            mean = sm.mean(self._returns, sample_weight=sample_weight)
            # Add new views and solve
            self._add_variance_views(mean=mean)
            sample_weight = self._solve_with_cvar()

        # Step 3: Mean, VaR, CVaR, Variance, Correlation, Skew and Kurtosis
        if (
            self.correlation_views is not None
            or self.skew_views is not None
            or self.kurtosis_views is not None
        ):
            # Get mean and variance from Step 2
            mean = sm.mean(self._returns, sample_weight=sample_weight)
            variance = sm.variance(
                self._returns, sample_weight=sample_weight, biased=True
            )
            # Add new views and solve
            self._add_correlation_views(mean=mean, variance=variance)
            self._add_skew_views(mean=mean, variance=variance)
            self._add_kurtosis_views(mean=mean, variance=variance)
            sample_weight = self._solve_with_cvar()

        self.relative_entropy_ = float(
            np.sum(scs.rel_entr(sample_weight, self._prior_sample_weight))
        )
        self.effective_number_of_scenarios_ = np.exp(sts.entropy(sample_weight))
        self.return_distribution_ = ReturnDistribution(
            mu=sm.mean(self._returns, sample_weight=sample_weight),
            covariance=np.cov(self._returns, rowvar=False, aweights=sample_weight),
            returns=self._returns,
            sample_weight=sample_weight,
        )

        # Manage memory
        del self._returns
        del self._constraints
        del self._prior_sample_weight

        return self

    def _add_constraint(self, a: np.ndarray, b: np.ndarray, name: str) -> None:
        """Add the left matrix `a` and right vector `b` of linear equality constraints
        `x @ a == b` and linear inequality constraints `x @ a <= b` to the
        `_constraints` dict.

        Parameters
        ----------
        a : ndarray of shape (n_observations, n_constraints)
            Left matrix in `x @ a == b` or `x @ a <= b`.

        a : ndarray of shape (n_observations, n_constraints)
            Right vector in `x @ a == b` or `x @ a <= b`.

        Returns
        -------
        None
        """
        if b.size == 0:
            return

        # Init constraints dict
        if self._constraints[name] is None:
            n_observations, _ = self._returns.shape
            self._constraints[name] = [np.empty((n_observations, 0)), np.empty(0)]

        # Rescaling: views can be on different scales, by rescaling we avoid high
        # disparity, have better conditioning, uniform stopping criteria and slack
        # penalties.
        scales = np.linalg.norm(a, axis=0)
        a /= scales
        b /= scales

        for i, x in enumerate([a, b]):
            self._constraints[name][i] = np.hstack((self._constraints[name][i], x))

    def _add_mean_views(self) -> None:
        """Add mean view constraints to the optimization problem."""
        if self.mean_views is None:
            return

        a_eq, b_eq, a_ineq, b_ineq = self._process_views(measure=PerfMeasure.MEAN)

        for a, b, name in [(a_eq, b_eq, "equality"), (a_ineq, b_ineq, "inequality")]:
            if a.size != 0:
                self._add_constraint(a=self._returns @ a.T, b=b, name=name)

    def _add_variance_views(self, mean: np.ndarray) -> None:
        """Add variance view constraints to the optimization problem.

        Parameters
        ----------
        mean : ndarray of shape (n_assets,)
            The fixed mean vector used to compute variance as a linear function of
            sample weights.
        """
        if self.variance_views is None:
            return

        a_eq, b_eq, a_ineq, b_ineq = self._process_views(measure=RiskMeasure.VARIANCE)

        _, n_assets = self._returns.shape

        fix = np.zeros(n_assets, dtype=bool)
        for a, b, name in [(a_eq, b_eq, "equality"), (a_ineq, b_ineq, "inequality")]:
            if a.size != 0:
                self._add_constraint(
                    a=(self._returns - mean) ** 2 @ a.T, b=b, name=name
                )
                fix |= np.any(a != 0, axis=0)

        self._fix_mean(fix=fix, mean=mean)

    def _add_skew_views(self, mean: np.ndarray, variance: np.ndarray) -> None:
        """Add skew view constraints to the optimization problem.

        Parameters
        ----------
        mean : ndarray of shape (n_assets,)
            The fixed mean vector used to compute skew as a linear function of sample
            weights.

        variance : ndarray of shape (n_assets,)
            The fixed variance vector used to compute skew as a linear function of
            sample weights.
        """
        if self.skew_views is None:
            return

        a_eq, b_eq, a_ineq, b_ineq = self._process_views(measure=ExtraRiskMeasure.SKEW)

        _, n_assets = self._returns.shape

        fix = np.zeros(n_assets, dtype=bool)
        for a, b, name in [(a_eq, b_eq, "equality"), (a_ineq, b_ineq, "inequality")]:
            if a.size != 0:
                self._add_constraint(
                    a=(self._returns**3 - mean**3 - 3 * mean * variance)
                    / variance**1.5
                    @ a.T,
                    b=b,
                    name=name,
                )
                fix |= np.any(a != 0, axis=0)

        # Fix the mean and the variance of the correlation views  in order for the
        # correlation to match exactly the views
        self._fix_mean(fix=fix, mean=mean)
        self._fix_variance(fix=fix, mean=mean, variance=variance)

    def _add_kurtosis_views(self, mean: np.ndarray, variance: np.ndarray) -> None:
        """Add kurtosis view constraints to the optimization problem.

        Parameters
        ----------
        mean : ndarray of shape (n_assets,)
           The fixed mean vector used to compute kurtosis as a linear function of
           sample weights.

        variance : ndarray of shape (n_assets,)
           The fixed variance vector used to compute kurtosis as a linear function of
           sample weights.
        """
        if self.kurtosis_views is None:
            return

        a_eq, b_eq, a_ineq, b_ineq = self._process_views(
            measure=ExtraRiskMeasure.KURTOSIS
        )

        _, n_assets = self._returns.shape
        fix = np.zeros(n_assets, dtype=bool)
        for a, b, name in [(a_eq, b_eq, "equality"), (a_ineq, b_ineq, "inequality")]:
            if a.size != 0:
                self._add_constraint(
                    a=(
                        (
                            self._returns**4
                            - 4 * mean * self._returns**3
                            + 6 * mean**2 * self._returns**2
                            - 3 * mean**4
                        )
                        / variance**2
                        @ a.T
                    ),
                    b=b,
                    name=name,
                )
                fix |= np.any(a != 0, axis=0)

        self._fix_mean(fix=fix, mean=mean)
        self._fix_variance(fix=fix, mean=mean, variance=variance)

    def _add_value_at_risk_views(self) -> None:
        """Add Value-at-Risk (VaR) view constraints to the optimization problem."""
        if self.value_at_risk_views is None:
            return

        a_eq, b_eq, a_ineq, b_ineq = self._process_views(
            measure=ExtraRiskMeasure.VALUE_AT_RISK
        )

        if (a_eq.size != 0 and np.any(np.count_nonzero(a_eq, axis=1) != 1)) or (
            a_ineq.size != 0 and np.any(np.count_nonzero(a_ineq, axis=1) != 1)
        ):
            raise ValueError(
                "You cannot mix multiple assets in a single Value-at-Risk view."
            )

        if np.any(b_eq < 0) | np.any(a_ineq * b_ineq[:, np.newaxis] < 0):
            raise ValueError("Value-at-Risk views must be strictly positive.")

        n_observations, _ = self._returns.shape

        for a, b, name in [(a_eq, b_eq, "equality"), (a_ineq, b_ineq, "inequality")]:
            for ai, bi in zip(a, b, strict=True):
                idx = np.where(self._returns[:, ai.astype(bool)].flatten() <= -abs(bi))[
                    0
                ]
                if idx.size == 0:
                    raise ValueError(
                        f"The Value-at-Risk view of {bi:0.3%} is excessively extreme. "
                        "Consider lowering the view or adjusting your prior "
                        "distribution to include more extreme values."
                    )
                sign = 1 if name == "equality" or bi > 0 else -1
                ai = np.zeros(n_observations)
                ai[idx] = 1.0
                self._add_constraint(
                    a=sign * ai.reshape(-1, 1),
                    b=sign * np.array([1 - self.value_at_risk_beta]),
                    name=name,
                )

    def _add_correlation_views(self, mean: np.ndarray, variance: np.ndarray) -> None:
        """Add correlation view constraints to the optimization problem.

        Parameters
        ----------
        mean : ndarray of shape (n_assets,)
           The fixed mean vector used to compute correlation as a linear function of
           sample weights.

        variance : ndarray of shape (n_assets,)
           The fixed variance vector used to compute correlation as a linear function of
           sample weights.
        """
        if self.correlation_views is None:
            return

        assets = self._groups[0]
        _, n_assets = self._returns.shape
        asset_to_index = {asset: i for i, asset in enumerate(assets)}
        try:
            views = []
            for view in self.correlation_views:
                res = _parse_correlation_view(view, assets=assets)
                expression = res["expression"]
                corr_view = expression["constant"]
                if "prior_assets" in expression:
                    i, j = (asset_to_index[a] for a in expression["prior_assets"])
                    corr_view += (
                        np.corrcoef(self._returns[:, i], self._returns[:, j])[0][1]
                        * expression["multiplier"]
                    )
                    corr_view = np.clip(corr_view, 0 + 1e-8, 1 - 1e-8)
                views.append(
                    (
                        (asset_to_index[a] for a in res["assets"]),
                        res["operator"],
                        corr_view,
                    )
                )
        except KeyError as e:
            raise ValueError(f"Asset {e.args[0]} is missing from the assets.") from None

        fix = np.zeros(n_assets, dtype=bool)
        for (i, j), op, corr_view in views:
            if not 0 <= corr_view <= 1:
                raise ValueError("Correlation views must be between 0 and 1.")
            ai = self._returns[:, i] * self._returns[:, j]
            bi = mean[i] * mean[j] + corr_view * np.sqrt(variance[i] * variance[j])
            sign = 1 if op in [operator.eq, operator.lt, operator.le] else -1
            self._add_constraint(
                a=sign * ai.reshape(-1, 1),
                b=sign * np.array([bi]),
                name="equality" if op == operator.eq else "inequality",
            )
            fix[[i, j]] = True

        self._fix_mean(fix=fix, mean=mean)
        self._fix_variance(fix=fix, mean=mean, variance=variance)

    def _solve_with_cvar(self) -> np.ndarray:
        """Solve the entropy pooling problem handling CVaR view constraints.

        CVaR view constraints cannot be directly expressed as linear functions of the
        posterior probabilities. Therefore, when CVaR views are present, the EP problem
        must be solved by recursively solving a series of convex programs that
        approximate the non-linear CVaR constraint.

        Our approach improves upon Meucci's algorithm [1]_ by formulating the problem
        in continuous space as a function of the dual variables etas (VaR levels)
        rather than searching over discrete tail sizes. This formulation not only
        handles the CVaR constraint more directly but also supports multiple CVaR views
        on different assets.

        Although the overall problem is convex in the dual variables etas, it remains
        non-smooth due to the presence of the positive-part operator in the CVaR
        definition. Consequently, we employ derivative-free optimization methods.
        Specifically, for a single CVaR view we use a one-dimensional root-finding
        method (Brent's method), and for the multidimensional case (supporting multiple
        CVaR views) we utilize Powell's method for derivative-free convex descent.

        Returns
        -------
        sample_weight : ndarray of shape (n_observations,)
            The updated probability vector satisfying all view constraints.

        References
        ----------
        .. [1] "Fully Flexible Extreme Views",
                Journal of Risk, Meucci, Ardia & Keel (2011)
        """
        if self.cvar_views is None:
            sample_weight = self._solve()
            return sample_weight

        a_eq, b_eq, _, b_ineq = self._process_views(measure=RiskMeasure.CVAR)

        if b_ineq.size != 0:
            raise ValueError(
                "CVaR view inequalities are not supported, use equalities views `==`"
            )

        if np.any(np.count_nonzero(a_eq, axis=1) != 1):
            raise ValueError("You cannot mix multiple assets in a single CVaR view")

        if np.any(b_eq < 0):
            raise ValueError("CVaR view must be strictly positive")

        n_observations, _ = self._returns.shape
        asset_returns = self._returns[:, np.where(a_eq.sum(axis=0) != 0)[0]].T
        views = b_eq
        n_views = len(views)

        min_ret = -np.min(asset_returns, axis=1)
        invalid_views = views >= min_ret
        if np.any(invalid_views):
            msg = ""
            for v, m in zip(views[invalid_views], min_ret[invalid_views], strict=True):
                msg += (
                    "The CVaR views of "
                    + ", ".join([f"{v:.2%}" for v in views[invalid_views]])
                    + f" is excessively extreme and cannot exceed {m:.2%} which is the "
                    "worst realization. Consider lowering the view or adjusting your "
                    "prior distribution to include more extreme values."
                )
            raise ValueError(msg)

        def func(etas: list[float]) -> tuple[np.ndarray, float]:
            """Solve the EP with CVaR constraints for a given list of etas (VaR levels).

            Parameters
            ----------
            etas : list[float]
                The list of etas (VaR levels) of each asset with a CVaR view.

            Returns
            -------
            sample_weight : ndarray of shape (n_observations,)
                Sample weight of the CVaR EP problem.
            error : float
                For one-dimensional (a single eta), the error is the difference
                between the target CVaR beta and the effective CVaR beta.
                For multidimensional, the error is the RMSE of the difference between
                the target CVaR and the effective CVaR.
            """
            # Init CVaR constraints
            self._constraints["cvar_equality"] = [
                np.empty((n_observations, 0)),
                np.empty(0),
            ]
            pos_part = None
            for i in range(n_views):
                if not (0 < etas[i] < views[i]):
                    raise ValueError(
                        f"eta[{i}] must be between 0 and the CVaR view {views[i]}"
                    )
                pos_part = np.maximum(-asset_returns[i] - etas[i], 0)
                self._add_constraint(
                    a=(pos_part / (1 - self.cvar_beta)).reshape(-1, 1),
                    b=np.array([views[i] - etas[i]]),
                    name="cvar_equality",
                )
            w = self._solve()

            if n_views == 1:
                error = np.sum(w[pos_part != 0]) - (1 - self.cvar_beta)
            else:
                error = np.linalg.norm(
                    sm.cvar(asset_returns.T, beta=self.cvar_beta, sample_weight=w)
                    - views
                ) / np.sqrt(n_views)
            return w, error

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="^Solution may be inaccurate")

            # One-dimensional: we use one-dimensional root-finding method
            if n_views == 1:
                sol = sco.root_scalar(
                    lambda x: func([x])[-1],
                    bracket=[1e-3, views[0] - 1e-8],
                    method="brentq",  # Doesn't required a smooth derivative
                    xtol=1e-4,
                    maxiter=50,
                )
                if not sol.converged:
                    raise RuntimeError(
                        "Failed to solve the CVaR view problem. Consider relaxing your "
                        "constraints or substituting with VaR views."
                    )
                res = [sol.root]
            # Multidimensional: we use derivative-free convex descent
            else:
                sol = sco.minimize(
                    lambda x: func(x)[-1],
                    x0=np.array([view * 0.5 for view in views]),
                    bounds=[(1e-3, view - 1e-8) for view in views],
                    method="Powell",
                    options={"xtol": 1e-4, "maxiter": 80},
                )
                if not sol.success:
                    raise ValueError(
                        "Failed to solve the multi-CVaR view problem. Consider "
                        "relaxing your constraints, using a single CVaR view, or "
                        "substituting with VaR views."
                    )
                res = sol.x

        sample_weight = func(res)[0]
        return sample_weight

    def _solve(self) -> np.ndarray:
        """Solve the base entropy pooling problem.
        Dispatch either to the EP dual (TNC) solver or the EP primal (CVXPY) solver
        based on the `solver` parameter.

        Returns
        -------
        sample_weight : ndarray of shape (n_observations,)
           The updated posterior probability vector.
        """
        if all(v is None for v in self._constraints.values()):
            # No view constraints so we don't need to solve the EP problem.
            return self._prior_sample_weight

        if self.solver == "TNC":
            return self._solve_dual()
        return self._solve_primal()

    def _solve_dual(self) -> np.ndarray:
        r"""Solves the entropic-pooling dual via SciPy's Truncated Newton
        Constrained method. By exploiting the smooth Fenchel dual and its
        closed-form gradient, it operates in :math:`\mathbb{R}^k` (the number of
        constraints) rather than :math:`\mathbb{R}^T` (the number of scenarios),
        yielding an order-of-magnitude speedup over primal CVXPY interior-point
        solvers.

        Returns
        -------
        sample_weight : ndarray of shape (n_observations,)
           The updated posterior probability vector.

        References
        ----------
        .. [1] "Convex Optimization", Cambridge University Press, Section 5.2.3,
            Entropy maximization, Boyd & Vandenberghe (2004)

        .. [2] "I-Divergence Geometry of Probability Distributions and Minimization
            Problems", The Annals of Probability, Csiszar (1975)

        .. [3] "Fully Flexible Views: Theory and Practice",
                Risk, Meucci (2013).

        """
        n_observations, _ = self._returns.shape
        # Init constraints with sum(p)==1, rescaled by its norm
        # Has better convergence than the normalized form done inside the dual.
        a = [np.ones(n_observations).reshape(-1, 1) / np.sqrt(n_observations)]
        b = [np.array([1.0]) / np.sqrt(n_observations)]
        bounds = [(None, None)]
        for name, constraints in self._constraints.items():
            if constraints is not None:
                a.append(constraints[0])
                b.append(constraints[1])
                s = constraints[1].size
                match name:
                    case "equality" | "cvar_equality":
                        bounds += [(None, None)] * s
                    case "fixed_equality":
                        # Equivalent to relaxing the problem with slack variables with
                        # a norm penalty to avoid solver infeasibility that may arise
                        # from overly tight constraints from fixing the moments.
                        bounds += [(-1000, 1000)] * s
                    case "inequality":
                        bounds += [(0, None)] * s
                    case _:
                        raise KeyError(f"constrain {name}")

        a = np.hstack(a)
        b = np.hstack(b)

        def func(x: np.ndarray) -> tuple[float, np.ndarray]:
            """Computes the Fenchel dual of the entropic-pooling problem in its
            unnormalized form.

            Parameters
            ----------
            x : ndarray of shape (n_constraints,)
               Dual variables (Lagrange multipliers for each constraint).

            Returns
            -------
            obj : float
               Value of the dual objective.
            grad : ndarray of shape  (n_constraints,)
               Gradient of the dual objective.
            """
            z = self._prior_sample_weight * np.exp(-a @ x - 1)
            obj = z.sum() + x @ b
            grad = b - a.T @ z
            return obj, grad

        # We use TNC as it often outperforms L-BFGS-B on EP problems because
        # it builds (implicitly) true curvature information via Hessian-vector products,
        # whereas L-BFGS-B only ever uses a low-rank quasi-Newton approximation.
        # We set stepmx=1 to avoid large solver defaults when the bounds are not None.
        solver_params = (
            self.solver_params
            if self.solver_params is not None
            else {
                "maxfun": 5000,
                "ftol": 1e-11,
                "xtol": 1e-8,
                "gtol": 1e-8,
                "stepmx": 1,
            }
        )

        sol = sco.minimize(
            func,
            x0=np.zeros_like(b),
            jac=True,
            bounds=bounds,
            method="TNC",
            options=solver_params,
        )
        if not sol.success:
            raise SolverError(
                "Dual problem with Solver 'TNC' failed. This typically occurs when the "
                "specified views conflict or are overly extreme. Consider using a "
                "prior that generates more synthetic data for extreme views. You can "
                "also change `solver_params` or try another `solver` such as "
                f"'CLARABEL'. Solver error: {sol.message}"
            )
        sample_weight = self._prior_sample_weight * np.exp(-1 - a @ sol.x)
        # Handles numerical precision errors
        sample_weight = np.clip(sample_weight, 0, 1)
        sample_weight /= sample_weight.sum()
        return sample_weight

    def _solve_primal(self) -> np.ndarray:
        """Solve the base entropy-pooling problem in its primal form by minimizing KL
        divergence to the prior.

        Returns
        -------
        sample_weight : ndarray of shape (n_observations,)
           The updated posterior probability vector.
        """
        n_observations, _ = self._returns.shape

        solver_params = self.solver_params if self.solver_params is not None else {}

        posterior = cp.Variable(n_observations)  # Posterior probas
        objective = cp.sum(cp.kl_div(posterior, self._prior_sample_weight))
        constraints = [posterior >= 0, cp.sum(posterior) == 1]

        if self._constraints["equality"] is not None:
            a, b = self._constraints["equality"]
            constraints.append(posterior @ a - b == 0)

        if self._constraints["inequality"] is not None:
            a, b = self._constraints["inequality"]
            constraints.append(posterior @ a - b <= 0)

        if self._constraints["cvar_equality"] is not None:
            a, b = self._constraints["cvar_equality"]
            constraints.append(posterior @ a - b == 0)

        if self._constraints["fixed_equality"] is not None:
            a, b = self._constraints["fixed_equality"]
            # Relax the problem with slack variables with a norm1 penalty to avoid
            # solver infeasibility that may arise from overly tight constraints from
            # fixing the moments.
            slack = cp.Variable(b.size)
            constraints.append(posterior @ a - b == slack)
            objective += 1e5 * cp.norm1(slack)

        problem = cp.Problem(cp.Minimize(objective), constraints)
        try:
            # We suppress cvxpy warning as it is redundant with our warning
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                problem.solve(solver=self.solver, **solver_params)

            if posterior.value is None:
                raise cp.SolverError
            if problem.status != cp.OPTIMAL:
                warnings.warn(
                    "Solution may be inaccurate. Try changing the solver params. For "
                    "more details, set `solver_params=dict(verbose=True)`",
                    stacklevel=2,
                )
            # Handles numerical precision errors
            sample_weight = np.clip(posterior.value, 0, 1)
            sample_weight /= sample_weight.sum()
            return sample_weight
        except (cp.SolverError, scl.ArpackNoConvergence):
            raise SolverError(
                f"Primal problem with Solver '{self.solver}' failed. This typically "
                "occurs when the specified views conflict or are overly extreme. "
                "Consider using a prior that generates more synthetic data for extreme "
                "views. You can also change `solver_params` or try another `solver` "
                "such as 'TNC'."
            ) from None

    def _process_views(self, measure: PerfMeasure | RiskMeasure | ExtraRiskMeasure):
        """Process and convert view equations into constraint matrices.

        This method uses the provided view strings and groups to generate the equality
        and inequality matrices (a_eq, b_eq, a_ineq, b_ineq) needed to formulate the
        constraints. Prior asset expressions are replaced using the current prior.

        Parameters
        ----------
        measure : {PerfMeasure, RiskMeasure, ExtraRiskMeasure}
           The type of view measure to process.

        Returns
        -------
        a_eq, b_eq, a_ineq, b_ineq : tuple of ndarray
           Matrices and vectors corresponding to equality and inequality constraints.
        """
        assets = self._groups[0]
        name = f"{measure.value}_views"
        views = getattr(self, name)
        views = np.asarray(views)
        if views.ndim != 1:
            raise ValueError(f"{name} must be a list of strings")
        measure_func = getattr(sm, str(measure.value))
        measure_args = {}
        if measure == RiskMeasure.VARIANCE:
            measure_args["biased"] = True
        required_prior_assets = _extract_prior_assets(views, assets=assets)
        if len(required_prior_assets) != 0:
            prior_values = {
                k: measure_func(self._returns[:, i], **measure_args)
                for i, k in enumerate(assets)
                if k in required_prior_assets
            }
            views = _replace_prior_views(views=views, prior_values=prior_values)

        a_eq, b_eq, a_ineq, b_ineq = equations_to_matrix(
            groups=self._groups,
            equations=views,
            raise_if_group_missing=True,
            names=("groups", "views"),
        )

        return a_eq, b_eq, a_ineq, b_ineq

    def _fix_mean(self, fix: np.ndarray, mean: np.ndarray) -> None:
        """Add constraints to fix the mean for assets where view constraints have been
        applied.

        The method introduces slack variables to avoid solver infeasibility from
        conflicting tight constraints.

        Parameters
        ----------
        fix : ndarray of bool of shape (n_assets,)
            Boolean mask indicating which assets to fix.

        mean : ndarray of shape (n_assets,)
            Fixed mean values for the assets.
        """
        fix &= ~self._is_fixed_mean
        if np.any(fix):
            self._add_constraint(
                a=self._returns[:, fix], b=mean[fix], name="fixed_equality"
            )
            self._is_fixed_mean |= fix

    def _fix_variance(
        self, fix: np.ndarray, mean: np.ndarray, variance: np.ndarray
    ) -> None:
        """Add constraints to fix the variance for assets where view constraints have
        been applied.

        The method introduces slack variables to avoid solver infeasibility from
        conflicting tight constraints.

        Parameters
        ----------
        fix : ndarray of bool of shape (n_assets,)
            Boolean mask indicating which assets to fix.

        mean : ndarray of shape (n_assets,)
            Fixed mean values used for the linearization of the variance.

        variance : np.ndarray
            Fixed variance values.
        """
        fix &= ~self._is_fixed_variance
        if np.any(fix):
            self._add_constraint(
                a=(self._returns[:, fix] - mean[fix]) ** 2,
                b=variance[fix],
                name="fixed_equality",
            )
            self._is_fixed_variance |= fix


def _extract_prior_assets(views: Sequence[str], assets: Sequence[str]) -> set[str]:
    """
    Given a list of views, return a set of asset names referenced within any
    'prior(ASSET)' pattern. Only asset names in the provided 'assets' list
    will be recognized.

    Supported format:
      - "prior(ASSET)"
      - "prior(ASSET) * a"
      - "a * prior(ASSET)"

    Parameters
    ----------
    views : list[str]
        The list of views to scan.

    assets : list[str]
        The list of allowed asset names.

    Returns
    -------
    prior_assets : set[str]
        Set of asset names that appear in prior() pattern.
    """
    allowed_assets = "|".join(re.escape(asset) for asset in assets)
    pattern = r"prior\(\s*(" + allowed_assets + r")\s*\)"

    prior_assets = set()
    for view in views:
        matches = re.findall(pattern, view)
        prior_assets.update(matches)

    return set(prior_assets)


def _replace_prior_views(
    views: Sequence[str], prior_values: dict[str, float]
) -> list[str]:
    """
    Replace occurrences of the below prior patterns using `prior_values`.

    Supported patterns:
        - "prior(ASSET)"
        - "prior(ASSET) * a"
        - "a * prior(ASSET)"

    Parameters
    ----------
    views : list[str]
        The list of views.

    prior_values : dict[str, float]
        A dictionary mapping asset names (str) to their prior values.

    Returns
    -------
    views : list[str]
        The views with each prior pattern instance replaced by its computed value.
    """
    # Build a regex pattern for allowed asset names.
    allowed_assets = "|".join(re.escape(asset) for asset in prior_values)

    # Pattern captures:
    #   - An optional multiplier before: ([0-9\.]+)\s*\*\s*
    #   - The prior() function with an allowed asset: prior\(\s*(allowed_asset)\s*\)
    #   - An optional multiplier after: \s*\*\s*([0-9\.]+)
    pattern = (
        r"(?:([0-9\.]+)\s*\*\s*)?"  # Optional pre-multiplier
        r"prior\(\s*(" + allowed_assets + r")\s*\)"  # prior(ASSET) with allowed asset
        r"(?:\s*\*\s*([0-9\.]+))?"  # Optional post-multiplier
    )

    def repl(match) -> str:
        pre_multiplier = float(match.group(1)) if match.group(1) else 1.0
        asset = match.group(2)
        post_multiplier = float(match.group(3)) if match.group(3) else 1.0
        result_value = prior_values[asset] * pre_multiplier * post_multiplier
        return str(result_value)

    new_views = [re.sub(pattern, repl, view) for view in views]

    # After substitution, check for any unresolved 'prior(ASSET)'
    unresolved_pattern = r"prior\(\s*([^)]+?)\s*\)"
    for view in new_views:
        m = re.search(unresolved_pattern, view)
        if m is not None:
            missing_asset = m.group(1)
            raise ValueError(
                "Unresolved 'prior' expression found in view. "
                f"Asset '{missing_asset}' is not available in prior_values."
            )

    return new_views


def _parse_correlation_view(view: str, assets: Sequence[str]) -> dict:
    """
    Parse a correlation view and return its structured representation.

    The view string should follow one of the formats:
      (asset1, asset2) <operator> <float>
      (asset1, asset2) <operator> <float> * prior(asset3, asset4) * <float> + <float>

    Only asset names from the provided `assets` list are accepted.
    If the view contains a "prior(...)" expression, it is replaced by its computed value
    using multipliers and constant additions. If the asset referenced in any prior
    expression is not allowed, an error is raised.

    Parameters
    ----------
    view : str
        The correlation view string.

    assets : list[str]
        A list of allowed asset names.

    Returns
    -------
    views : dict
        A dictionary with keys:
         - "assets": tuple(asset1, asset2)
         - "operator": the comparison operator (==, >=, <=, >, or <)
         - "expression": a dict representing the right-hand side expression.
             * If there is no prior expression, "expression" contains {"constant": value}.
             * If a prior expression is present, it returns a dict with keys:
                   "prior_assets": tuple(asset3, asset4),
                   "multiplier": overall multiplier (default 1.0),
                   "constant": constant term (default 0.0).
    """
    operator_map = {
        ">": operator.gt,
        ">=": operator.ge,
        "<": operator.lt,
        "<=": operator.le,
        "==": operator.eq,
    }

    # Build a regex pattern that only accepts the allowed asset names.
    allowed_assets = "|".join(re.escape(asset) for asset in assets)

    main_pattern = (
        r"^\(\s*(" + allowed_assets + r")\s*,\s*(" + allowed_assets + r")\s*\)\s*"
        r"(==|>=|<=|>|<)\s*(.+)$"
    )
    main_match = re.match(main_pattern, view)
    if not main_match:
        raise ValueError(
            f"Invalid correlation view format or unknown asset in view: {view}"
        )

    asset1, asset2, operator_str, expression = main_match.groups()
    expression = expression.strip()

    if "prior(" not in expression:
        try:
            constant = float(expression)
        except ValueError as e:
            raise ValueError(
                f"Could not convert constant '{expression}' to float in view: {view}"
            ) from e
        parsed_expression = {"constant": constant}
    else:
        # Pattern to capture:
        #   - An optional pre multiplier followed by "*" (e.g. "2 *")
        #   - The pattern prior( asset3 , asset4 )
        #   - An optional post multiplier preceded by "*" (e.g. "* 3")
        #   - An optional constant preceded by "+" (e.g. "+ 4")
        prior_pattern = (
            r"^(?:\s*([A-Za-z0-9_.-]+)\s*\*\s*)?"  # Optional pre-multiplier
            r"prior\(\s*("
            + allowed_assets
            + r")\s*,\s*("
            + allowed_assets
            + r")\s*\)"  # prior(asset3, asset4)
            r"(?:\s*\*\s*([A-Za-z0-9_.-]+))?"  # Optional post-multiplier
            r"(?:\s*\+\s*([A-Za-z0-9_.-]+))?"  # Optional constant addition
            r"$"
        )
        prior_match = re.match(prior_pattern, expression)
        if not prior_match:
            raise ValueError(
                "Invalid prior expression format or unknown asset in expression: "
                f"{expression}"
            )

        (pre_mult, asset3, asset4, post_mult, constant) = prior_match.groups()
        try:
            pre_mult = float(pre_mult) if pre_mult is not None else 1.0
        except ValueError as e:
            raise ValueError(
                f"Invalid pre-multiplier '{pre_mult}' in view: {view}"
            ) from e
        try:
            post_mult = float(post_mult) if post_mult is not None else 1.0
        except ValueError as e:
            raise ValueError(
                f"Invalid post-multiplier '{post_mult}' in view: {view}"
            ) from e
        try:
            constant = float(constant) if constant is not None else 0.0
        except ValueError as e:
            raise ValueError(f"Invalid constant '{constant}' in view: {view}") from e

        parsed_expression = {
            "prior_assets": (asset3, asset4),
            "multiplier": pre_mult * post_mult,
            "constant": constant,
        }

    return {
        "assets": (asset1, asset2),
        "operator": operator_map[operator_str],
        "expression": parsed_expression,
    }



"""Johnson SU Estimator."""

# Copyright (c) 2025
# Authors: The skfolio developers
# Credits: Matteo Manzi, Vincent Maladière, Carlo Nicolini
# SPDX-License-Identifier: BSD-3-Clause

import numpy.typing as npt
import scipy.stats as st

from skfolio.distribution.univariate._base import BaseUnivariateDist


class JohnsonSU(BaseUnivariateDist):
    r"""Johnson SU Distribution Estimation.

    This estimator fits a univariate Johnson SU distribution to the input data.
    The Johnson SU distribution is flexible and can capture both skewness and fat tails,
    making it appropriate for financial time series modeling.

    The probability density function is:

    .. math::

        f(x, a, b) = \frac{b}{\sqrt{x^2 + 1}}
                     \phi(a + b \log(x + \sqrt{x^2 + 1}))

    where :math:`x`, :math:`a`, and :math:`b` are real scalars; :math:`b > 0`.
    :math:`\phi` is the pdf of the normal distribution.

    The probability density above is defined in the "standardized" form. To shift
    and/or scale the distribution use the loc and scale parameters. Specifically,
    `pdf(x, a, b, loc, scale)` is equivalent to `pdf(y, a, b) / scale` with
    `y = (x - loc) / scale`.

    For more information, you can refer to the `scipy documentation <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.johnsonsu.html#scipy.stats.johnsonsu>`_

    Parameters
    ----------
    loc : float, optional
        If provided, the location parameter is fixed to this value during fitting.
        Otherwise, it is estimated from the data.

    scale : float, optional
        If provided, the scale parameter is fixed to this value during fitting.
        Otherwise, it is estimated from the data.

    random_state : int, RandomState instance or None, default=None
        Seed or random state to ensure reproducibility.

    Attributes
    ----------
    a_ : float
        The fitted first shape parameter of the Johnson SU distribution.

    b_ : float
        The fitted second shape parameter of the Johnson SU distribution.

    loc_ : float
        The fitted location parameter.

    scale_ : float
        The fitted scale parameter.

    Examples
    --------
    >>> from skfolio.datasets import load_sp500_index
    >>> from skfolio.preprocessing import prices_to_returns
    >>> from skfolio.distribution.univariate import JohnsonSU
    >>>
    >>> # Load historical prices and convert them to returns
    >>> prices = load_sp500_index()
    >>> X = prices_to_returns(prices)
    >>>
    >>> # Initialize the estimator.
    >>> model = JohnsonSU()
    >>>
    >>> # Fit the model to the data.
    >>> model.fit(X)
    >>>
    >>> # Display the fitted parameters.
    >>> print(model.fitted_repr)
    JohnsonSU(0.0742, 1.08, 0.00115, 0.00774)
    >>>
    >>> # Compute the log-likelihood, total log-likelihood, CDF, PPF, AIC, and BIC
    >>> log_likelihood = model.score_samples(X)
    >>> score = model.score(X)
    >>> cdf = model.cdf(X)
    >>> ppf = model.ppf(X)
    >>> aic = model.aic(X)
    >>> bic = model.bic(X)
    >>>
    >>> # Generate 5 new samples from the fitted distribution.
    >>> samples = model.sample(n_samples=5)
    >>>
    >>> # Plot the estimated probability density function (PDF).
    >>> fig = model.plot_pdf()
    >>> fig.show()
    """

    a_: float
    b_: float
    loc_: float
    scale_: float
    _scipy_model = st.johnsonsu

    def __init__(
        self,
        loc: float | None = None,
        scale: float | None = None,
        random_state: int | None = None,
    ):
        super().__init__(random_state=random_state)
        self.loc = loc
        self.scale = scale

    @property
    def _scipy_params(self) -> dict[str, float]:
        """Dictionary of parameters to pass to the underlying SciPy distribution."""
        return {"a": self.a_, "b": self.b_, "loc": self.loc_, "scale": self.scale_}

    def fit(self, X: npt.ArrayLike, y=None) -> "JohnsonSU":
        """Fit the univariate Johnson SU distribution model.

        Parameters
        ----------
        X : array-like of shape (n_observations, 1)
            The input data. X must contain a single column.

        y : None
            Ignored. Provided for compatibility with scikit-learn's API.

        Returns
        -------
        self : JohnsonSU
            Returns the instance itself.
        """
        X = self._validate_X(X, reset=True)

        if self.loc is not None and self.scale is not None:
            raise ValueError("Either loc or scale must be None to be fitted")

        fixed_params = {}
        if self.loc is not None:
            fixed_params["floc"] = self.loc
        if self.scale is not None:
            fixed_params["fscale"] = self.scale

        self.a_, self.b_, self.loc_, self.scale_ = self._scipy_model.fit(
            X, **fixed_params
        )
        return self



"""Equilibrium Expected Returns (Mu) Estimators."""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause
# Implementation derived from:
# scikit-learn, Copyright (c) 2007-2010 David Cournapeau, Fabian Pedregosa, Olivier
# Grisel Licensed under BSD 3 clause.

import numpy as np
import numpy.typing as npt
import sklearn.utils.metadata_routing as skm
import sklearn.utils.validation as skv

from skfolio.moments.covariance import BaseCovariance, EmpiricalCovariance
from skfolio.moments.expected_returns._base import BaseMu
from skfolio.utils.tools import check_estimator


class EquilibriumMu(BaseMu):
    r"""Equilibrium Expected Returns (Mu) estimator.

    The Equilibrium is defined as:

        .. math:: risk\_aversion \times \Sigma \cdot w^T

    For Market Cap Equilibrium, the weights are the assets Market Caps.
    For Equal-weighted Equilibrium, the weights are equal-weighted (1/N).

    Parameters
    ----------
    risk_aversion : float, default=1.0
        Risk aversion factor.
        The default value is `1.0`.

    weights : array-like of shape (n_assets,), optional
        Asset weights used to compute the Expected Return Equilibrium.
        The default is to use the equal-weighted equilibrium (1/N).
        For a Market Cap weighted equilibrium, you must provide the asset Market Caps.

    covariance_estimator : BaseCovariance, optional
        :ref:`Covariance estimator <covariance_estimator>` used to estimate the
        covariance in the equilibrium formula.
        The default (`None`) is to use :class:`~skfolio.moments.EmpiricalCovariance`.

    Attributes
    ----------
    mu_ : ndarray of shape (n_assets,)
          Estimated expected returns of the assets.

    covariance_estimator_ : BaseCovariance
        Fitted `covariance_estimator`.

    n_features_in_ : int
        Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of assets seen during `fit`. Defined only when `X`
        has assets names that are all strings.
    """

    covariance_estimator_: BaseCovariance

    def __init__(
        self,
        risk_aversion: float = 1,
        weights: np.ndarray | None = None,
        covariance_estimator: BaseCovariance | None = None,
    ):
        self.risk_aversion = risk_aversion
        self.weights = weights
        self.covariance_estimator = covariance_estimator

    def get_metadata_routing(self):
        # noinspection PyTypeChecker
        router = skm.MetadataRouter(owner=self.__class__.__name__).add(
            covariance_estimator=self.covariance_estimator,
            method_mapping=skm.MethodMapping().add(caller="fit", callee="fit"),
        )
        return router

    def fit(self, X: npt.ArrayLike, y=None, **fit_params) -> "EquilibriumMu":
        """Fit the EquilibriumMu estimator model.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
           Price returns of the assets.

        y : Ignored
            Not used, present for API consistency by convention.

        **fit_params : dict
            Parameters to pass to the underlying estimators.
            Only available if `enable_metadata_routing=True`, which can be
            set by using ``sklearn.set_config(enable_metadata_routing=True)``.
            See :ref:`Metadata Routing User Guide <metadata_routing>` for
            more details.

        Returns
        -------
        self : EquilibriumMu
            Fitted estimator.
        """
        routed_params = skm.process_routing(self, "fit", **fit_params)

        # fitting estimators
        self.covariance_estimator_ = check_estimator(
            self.covariance_estimator,
            default=EmpiricalCovariance(),
            check_type=BaseCovariance,
        )
        # noinspection PyArgumentList
        self.covariance_estimator_.fit(X, y, **routed_params.covariance_estimator.fit)

        # we validate and convert to numpy after all models have been fitted to keep
        # features names information.
        X = skv.validate_data(self, X)
        n_assets = X.shape[1]
        if self.weights is None:
            weights = np.ones(n_assets) / n_assets
        else:
            weights = np.asarray(self.weights)
        self.mu_ = self.risk_aversion * self.covariance_estimator_.covariance_ @ weights
        return self



"""Bivariate Gumbel Copula Estimation."""

# Copyright (c) 2025
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# Credits: Matteo Manzi, Vincent Maladière, Carlo Nicolini
# SPDX-License-Identifier: BSD-3-Clause

import numpy as np
import numpy.typing as npt
import scipy.stats as st
import sklearn.utils.validation as skv

from skfolio.distribution.copula._base import BaseBivariateCopula
from skfolio.distribution.copula._utils import (
    CopulaRotation,
    _apply_copula_rotation,
    _apply_margin_swap,
    _apply_rotation_cdf,
    _apply_rotation_partial_derivatives,
    _select_rotation_itau,
    _select_theta_and_rotation_mle,
)

# Gumbel copula with a theta of 1.0 is just the independence copula, so we chose a lower
# bound of 1.0001. After 80, the copula is already imposing extremely strong upper tail
# dependence and increasing it will make it impractical.
_THETA_BOUNDS = (1.0001, 80.0)


class GumbelCopula(BaseBivariateCopula):
    r"""Bivariate Gumbel Copula Estimation.

    The Gumbel copula is an Archimedean copula characterized by strong upper tail
    dependence and little to no lower tail dependence.

    In its unrotated form, it is used for modeling extreme co-movements in the upper
    tail (i.e. simultaneous extreme gains).

    Rotations allow the copula to be adapted for different types of tail dependence:
      - A 180° rotation captures extreme co-movements in the lower tail (i.e.
        simultaneous extreme losses).

      - A 90° rotation captures scenarios where one variable exhibits extreme losses
        while the other shows extreme gains.

      - A 270° rotation captures the opposite scenario, where one variable experiences
        extreme gains while the other suffers extreme losses.

    Gumbel copula generally exhibits weaker upper tail dependence than the Joe copula.

    It is defined by:

    .. math::
            C_{\theta}(u, v) = \exp\Bigl(-\Bigl[(-\ln u)^{\theta}+(-\ln v)^{\theta}\Bigr]^{1/\theta}\Bigr)

    where :math:`\theta \ge 1` is the dependence parameter. When :math:`\theta = 1`,
    the Gumbel copula reduces to the independence copula. Larger values of
    :math:`\theta` result in stronger upper-tail dependence.

    .. note::

        Rotations are needed for Archimedean copulas (e.g., Joe, Gumbel, Gumbel)
        because their parameters only model positive dependence, and they exhibit
        asymmetric tail behavior. To model negative dependence, one uses rotations
        to "flip" the copula's tail dependence.

    Parameters
    ----------
    itau : bool, default=True
        If True, :math:`\theta` is estimated using the Kendall's tau inversion method;
        otherwise, the Maximum Likelihood Estimation (MLE) method is used. The MLE is
        slower but more accurate.

    kendall_tau : float, optional
        If `itau` is True and `kendall_tau` is provided, this value is used;
        otherwise, it is computed.

    tolerance : float, default=1e-4
        Convergence tolerance for the MLE optimization.

    random_state : int, RandomState instance or None, default=None
        Seed or random state to ensure reproducibility.

    Attributes
    ----------
    theta_ : float
        Fitted theta coefficient :math:`\theta` > 1.

    rotation_ : CopulaRotation
        Fitted rotation of the copula.

    Examples
    --------
    >>> from skfolio.datasets import load_sp500_dataset
    >>> from skfolio.preprocessing import prices_to_returns
    >>> from skfolio.distribution import GumbelCopula, compute_pseudo_observations
    >>>
    >>> # Load historical prices and convert them to returns
    >>> prices = load_sp500_dataset()
    >>> X = prices_to_returns(prices)
    >>> X = X[["AAPL", "JPM"]]
    >>>
    >>> # Convert returns to pseudo observation in the interval [0,1]
    >>> X = compute_pseudo_observations(X)
    >>>
    >>> # Initialize the Copula estimator
    >>> model = GumbelCopula()
    >>>
    >>> # Fit the model to the data.
    >>> model.fit(X)
    >>>
    >>> # Display the fitted parameter and tail dependence coefficients
    >>> print(model.fitted_repr)
    GumbelCopula(theta=1.27, rot=180°)
    >>> print(model.lower_tail_dependence)
    0.2735
    >>> print(model.upper_tail_dependence)
    0.0
    >>>
    >>> # Compute the log-likelihood, total log-likelihood, CDF, Partial Derivative,
    >>> # Inverse Partial Derivative, AIC, and BIC
    >>> log_likelihood = model.score_samples(X)
    >>> score = model.score(X)
    >>> cdf = model.cdf(X)
    >>> p = model.partial_derivative(X)
    >>> u = model.inverse_partial_derivative(X)
    >>> aic = model.aic(X)
    >>> bic = model.bic(X)
    >>>
    >>> # Generate 5 new samples
    >>> samples = model.sample(n_samples=5)
    >>>
    >>> # Plot the tail concentration function.
    >>> fig = model.plot_tail_concentration()
    >>> fig.show()
    >>>
    >>> # Plot a 2D contour of the estimated PDF.
    >>> fig = model.plot_pdf_2d()
    >>> fig.show()
    >>>
    >>> # Plot a 3D surface of the estimated PDF.
    >>> fig = model.plot_pdf_3d()
    >>> fig.show()

    References
    ----------
    .. [1] "An Introduction to Copulas (2nd ed.)",
       Nelsen (2006)

    .. [2] "Multivariate Models and Dependence Concepts",
        Joe, Chapman & Hall (1997)

    .. [3] "Quantitative Risk Management: Concepts, Techniques and Tools",
        McNeil, Frey & Embrechts (2005)

    .. [4] "The t Copula and Related Copulas",
        Demarta & McNeil (2005)

    .. [5] "Copula Methods in Finance",
        Cherubini, Luciano & Vecchiato (2004)
    """

    theta_: float
    rotation_: CopulaRotation
    _n_params = 1

    def __init__(
        self,
        itau: bool = True,
        kendall_tau: float | None = None,
        tolerance: float = 1e-4,
        random_state: int | None = None,
    ):
        super().__init__(random_state=random_state)
        self.itau = itau
        self.kendall_tau = kendall_tau
        self.tolerance = tolerance

    def fit(self, X: npt.ArrayLike, y=None) -> "GumbelCopula":
        r"""Fit the Bivariate Gumbel Copula.

        If `itau` is True, estimates :math:`\theta` using Kendall's tau inversion.
        Otherwise, uses MLE by maximizing the log-likelihood.

        Parameters
        ----------
        X : array-like of shape (n_observations, 2)
            An array of bivariate inputs `(u, v)` where each row represents a
            bivariate observation. Both `u` and `v` must be in the interval [0, 1],
            having been transformed to uniform marginals.

        y : None
            Ignored. Provided for compatibility with scikit-learn's API.

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        X = self._validate_X(X, reset=True)

        if self.itau:
            if self.kendall_tau is None:
                kendall_tau = st.kendalltau(X[:, 0], X[:, 1]).statistic
            else:
                kendall_tau = self.kendall_tau

            # For Gumbel, the theoretical relationship is: tau = 1 - 1/theta.
            abs_kendall_tau = min(abs(kendall_tau), 0.9999)

            self.theta_ = np.clip(
                1.0 / (1.0 - abs_kendall_tau),
                a_min=_THETA_BOUNDS[0],
                a_max=_THETA_BOUNDS[1],
            )
            self.rotation_ = _select_rotation_itau(
                func=_neg_log_likelihood, X=X, theta=self.theta_
            )

        else:
            self.theta_, self.rotation_ = _select_theta_and_rotation_mle(
                _neg_log_likelihood, X=X, bounds=_THETA_BOUNDS, tolerance=self.tolerance
            )

        return self

    def cdf(self, X: npt.ArrayLike) -> np.ndarray:
        r"""Compute the CDF of the bivariate Gumbel copula.

        .. math::
            C(u,v) = \exp\Bigl(-\Bigl[(-\ln u)^{\theta}+(-\ln v)^{\theta}\Bigr]^{1/\theta}\Bigr).

        Parameters
        ----------
        X : array-like of shape (n_observations, 2)
            An array of bivariate inputs `(u, v)` where each row represents a
            bivariate observation. Both `u` and `v` must be in the interval [0, 1],
            having been transformed to uniform marginals.

        Returns
        -------
        cdf : ndarray of shape (n_observations,)
            CDF values for each observation in X.
        """
        skv.check_is_fitted(self)
        X = self._validate_X(X, reset=False)
        cdf = _apply_rotation_cdf(
            func=_base_cdf, X=X, rotation=self.rotation_, theta=self.theta_
        )
        return cdf

    def partial_derivative(
        self, X: npt.ArrayLike, first_margin: bool = False
    ) -> np.ndarray:
        r"""Compute the h-function (partial derivative) for the bivariate Gumbel copula
        with respect to a specified margin.

        The h-function with respect to the second margin represents the conditional
        distribution function of :math:`u` given :math:`v`:

        .. math::  \begin{aligned}
                   h(u \mid v)
                     &= \frac{\partial C(u,v)}{\partial v}\\[6pt]
                     &= C(u,v)\,\Bigl[(-\ln u)^{\theta}+(-\ln v)^{\theta}\Bigr]^{\frac{1}{\theta}-1}
                         \,(-\ln v)^{\theta-1}\,\frac{1}{v}.
                   \end{aligned}

        Parameters
        ----------
        X : array-like of shape (n_observations, 2)
            An array of bivariate inputs `(u, v)` where each row represents a
            bivariate observation. Both `u` and `v` must be in the interval [0, 1],
            having been transformed to uniform marginals.

        first_margin : bool, default=False
            If True, compute the partial derivative with respect to the first
            margin `u`; otherwise, compute the partial derivative with respect to the
            second margin `v`.

        Returns
        -------
        p  : ndarray of shape (n_observations,)
            h-function values :math:`h(u \mid v)` for each observation in X.
        """
        skv.check_is_fitted(self)
        X = self._validate_X(X, reset=False)
        p = _apply_rotation_partial_derivatives(
            func=_base_partial_derivative,  # Computation remains unchanged
            X=X,
            rotation=self.rotation_,
            first_margin=first_margin,
            theta=self.theta_,
        )
        return p

    def inverse_partial_derivative(
        self, X: npt.ArrayLike, first_margin: bool = False
    ) -> np.ndarray:
        r"""Compute the inverse of the bivariate copula's partial derivative, commonly
        known as the inverse h-function.

        Let :math:`C(u, v)` be a bivariate copula. The h-function with respect to the
        second margin is defined by

        .. math::
            h(u \mid v) \;=\; \frac{\partial\,C(u, v)}{\partial\,v},

        which is the conditional distribution of :math:`U` given :math:`V = v`.
        The **inverse h-function**, denoted :math:`h^{-1}(p \mid v)`, is the unique
        value :math:`u \in [0,1]` such that

        .. math::
            h(u \mid v) \;=\; p,
            \quad \text{where } p \in [0,1].

        In practical terms, given :math:`(p, v)` in :math:`[0, 1]^2`,
        :math:`h^{-1}(p \mid v)` solves for the :math:`u` satisfying
        :math:`p = \partial C(u, v)/\partial v`.

        Parameters
        ----------
        X : array-like of shape (n_observations, 2)
            An array of bivariate inputs `(p, v)`, each in the interval [0, 1].
            - The first column `p` corresponds to the value of the h-function.
            - The second column `v` is the conditioning variable.

        first_margin : bool, default=False
            If True, compute the inverse partial derivative with respect to the first
            margin `u`; otherwise, compute the inverse partial derivative with respect
            to the second margin `v`.

        Returns
        -------
        u : ndarray of shape (n_observations,)
            A 1D-array where each element is the computed :math:`u = h^{-1}(p \mid v)`
            for the corresponding pair in `X`.
        """
        skv.check_is_fitted(self)
        X = self._validate_X(X, reset=False)
        u = _apply_rotation_partial_derivatives(
            func=_base_inverse_partial_derivative,  # Computation remains unchanged
            X=X,
            rotation=self.rotation_,
            first_margin=first_margin,
            theta=self.theta_,
        )
        return u

    def score_samples(self, X: npt.ArrayLike) -> np.ndarray:
        r"""Compute the log-likelihood of each sample (log-pdf) under the model.

        For Gumbel, the PDF is given by:

        .. math::
            c(u,v) = \exp\Bigl(-\Bigl((-\ln u)^{\theta}+(-\ln v)^{\theta}\Bigr)^{1/\theta}\Bigr)
                     \cdot \left[\Bigl((-\ln u)^{\theta}+(-\ln v)^{\theta}\Bigr)^{1/\theta-1}
                          \left\{ \frac{(-\ln u)^{\theta}}{u}+\frac{(-\ln v)^{\theta}}{v}\right\}\right].

        Parameters
        ----------
        X : array-like of shape (n_observations, 2)
            An array of bivariate inputs `(u, v)` where each row represents a
            bivariate observation. Both `u` and `v` must be in the interval [0, 1],
            having been transformed to uniform marginals.

        Returns
        -------
        density : ndarray of shape (n_observations,)
            The log-likelihood of each sample under the fitted copula.
        """
        skv.check_is_fitted(self)
        X = self._validate_X(X, reset=False)
        X = _apply_copula_rotation(X, rotation=self.rotation_)
        log_density = _base_sample_scores(X=X, theta=self.theta_)
        return log_density

    @property
    def lower_tail_dependence(self) -> float:
        """Theoretical lower tail dependence coefficient."""
        skv.check_is_fitted(self)
        if self.rotation_ == CopulaRotation.R180:
            return 2.0 - np.power(2.0, 1.0 / self.theta_)
        return 0

    @property
    def upper_tail_dependence(self) -> float:
        """Theoretical upper tail dependence coefficient."""
        skv.check_is_fitted(self)
        if self.rotation_ == CopulaRotation.R0:
            return 2.0 - np.power(2.0, 1.0 / self.theta_)
        return 0

    @property
    def fitted_repr(self) -> str:
        """String representation of the fitted copula."""
        return (
            f"{self.__class__.__name__}(theta={self.theta_:0.2f}, rot={self.rotation_})"
        )


def _neg_log_likelihood(theta: float, X: np.ndarray) -> float:
    """Negative log-likelihood function for the Gumbel copula.

    Parameters
    ----------
    X : array-like of shape (n_observations, 2)
        An array of bivariate inputs `(u, v)` where each row represents a
        bivariate observation. Both `u` and `v` must be in the interval [0, 1],
        having been transformed to uniform marginals.

    theta : float
        The dependence parameter (must be greater than 1).

    Returns
    -------
    value : float
        The negative log-likelihood value.
    """
    return -np.sum(_base_sample_scores(X=X, theta=theta))


def _base_sample_scores(X: np.ndarray, theta: float) -> np.ndarray:
    r"""Compute the log-likelihood of each sample (log-pdf) under the bivariate Gumbel
    copula.

    Parameters
    ----------
    X : array-like of shape (n_observations, 2)
         Bivariate samples `(u, v)`, with each component in [0,1].

    theta : float
         The dependence parameter (must be greater than 1).

    Returns
    -------
    logpdf : ndarray of shape (n_observations,)
         Log-likelihood values for each observation.
    """
    if theta <= 1:
        raise ValueError("Theta must be greater than 1 for the Gumbel copula.")
    Z = -np.log(X)
    s = np.power(np.power(Z, theta).sum(axis=1), 1 / theta)
    s = np.clip(s, a_min=1e-10, a_max=None)
    log_density = (
        -s
        + np.log(s + theta - 1)
        + (1 - 2 * theta) * np.log(s)
        + (theta - 1) * np.log(Z.prod(axis=1))
        + Z.sum(axis=1)
    )
    return log_density


def _base_cdf(X: np.ndarray, theta: float) -> np.ndarray:
    r"""Bivariate Gumbel CDF (unrotated).

    .. math::
        C(u,v) = \exp\Bigl(-\Bigl[(-\ln u)^{\theta}+(-\ln v)^{\theta}\Bigr]^{1/\theta}\Bigr).
    """
    cdf = np.exp(-np.power(np.power(-np.log(X), theta).sum(axis=1), 1.0 / theta))
    return cdf


def _base_partial_derivative(
    X: np.ndarray, first_margin: bool, theta: float
) -> np.ndarray:
    r"""
    Compute the partial derivative (h-function) for the unrotated Gumbel copula.

    For Gumbel, the copula is defined as:

    .. math::
        C(u,v)=\exp\Bigl(-\Bigl[(-\ln u)^{\theta}+(-\ln v)^{\theta}\Bigr]^{1/\theta}\Bigr).

    The partial derivative with respect to v is:

    .. math::
        \frac{\partial C(u,v)}{\partial v}
          = C(u,v)\,\Bigl[(-\ln u)^{\theta}+(-\ln v)^{\theta}\Bigr]^{\frac{1}{\theta}-1}
            \,(-\ln v)^{\theta-1}\,\frac{1}{v}.

    Parameters
    ----------
    X : array-like of shape (n_observations, 2)
         An array of bivariate inputs `(u, v)` with values in [0, 1].

    first_margin : bool, default=False
         If True, compute with respect to u (by swapping margins); otherwise,
         compute with respect to v.

    theta : float
         The dependence parameter (must be > 1).

    Returns
    -------
    p : ndarray of shape (n_observations,)
         The computed h-function values.
    """
    X = _apply_margin_swap(X, first_margin=first_margin)
    _, v = X.T
    x, y = -np.log(X).T
    p = (
        np.exp(-np.power(np.power(x, theta) + np.power(y, theta), 1.0 / theta))
        * np.power(np.power(x / y, theta) + 1.0, 1.0 / theta - 1.0)
        / v
    )
    return p


def _base_inverse_partial_derivative(
    X: np.ndarray, first_margin: bool, theta: float
) -> np.ndarray:
    r"""
    Compute the inverse partial derivative for the unrotated Gumbel copula,
    i.e. solve for u in h(u|v)=p.

    In other words, given
      - p, the value of the h-function, and
      - v, the conditioning variable,
    solve:

    .. math::
      p = C(u,v)\,\Bigl[(-\ln u)^{\theta}+(-\ln v)^{\theta}\Bigr]^{\frac{1}{\theta}-1}\,
          (-\ln v)^{\theta-1}\,\frac{1}{v},

    for u ∈ [0,1]. Since no closed-form solution exists, we use a numerical method.

    Parameters
    ----------
    X : array-like of shape (n_observations, 2)
         An array with first column p (h-function values) and second column v
         (conditioning variable).

    first_margin : bool, default=False
         If True, treat the first margin as the conditioning variable.

    theta : float
         The dependence parameter (must be > 1).

    Returns
    -------
    u : ndarray of shape (n_observations,)
         A 1D-array where each element is the solution u ∈ [0,1] such that h(u|v)=p.
    """
    X = _apply_margin_swap(X, first_margin=first_margin)
    p, v = -np.log(X).T
    s = v + p + np.log(v) * (theta - 1.0)
    # Initial guess
    x = v.copy()
    max_iters = 50
    tol = 1e-8
    for _ in range(max_iters):
        x_new = x * (s - (theta - 1.0) * (np.log(x) - 1.0)) / (theta + x - 1.0)
        x_new = np.clip(x_new, x, None)
        diff = np.max(np.abs(x_new - x))
        x = x_new
        if diff < tol:
            break
    u = np.exp(-np.power(np.power(x, theta) - np.power(v, theta), 1.0 / theta))
    return u



"""Bivariate Independent Copula Estimation."""

# Copyright (c) 2025
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# Credits: Matteo Manzi, Vincent Maladière, Carlo Nicolini
# SPDX-License-Identifier: BSD-3-Clause

import numpy as np
import numpy.typing as npt
import sklearn.utils.validation as skv

from skfolio.distribution.copula._base import BaseBivariateCopula


class IndependentCopula(BaseBivariateCopula):
    r"""Bivariate Independent Copula (also called the product copula).

    It is defined by:

    .. math::
        C(u, v) = u \cdot v

    Parameters
    ----------
    random_state : int, RandomState instance or None, default=None
        Seed or random state to ensure reproducibility.

    References
    ----------
    .. [1] "An Introduction to Copulas (2nd ed.)",
       Nelsen (2006)

    .. [2] "Multivariate Models and Dependence Concepts",
        Joe, Chapman & Hall (1997)

    .. [3] "Quantitative Risk Management: Concepts, Techniques and Tools",
        McNeil, Frey & Embrechts (2005)

    .. [4] "The t Copula and Related Copulas",
        Demarta & McNeil (2005)

    .. [5] "Copula Methods in Finance",
        Cherubini, Luciano & Vecchiato (2004)
    """

    _n_params = 0

    def __init__(self, random_state: int | None = None):
        super().__init__(random_state=random_state)

    def fit(self, X: npt.ArrayLike, y=None) -> "IndependentCopula":
        """Fit the Bivariate Independent Copula.

        Provided for compatibility with the API.

        Parameters
        ----------
        X : array-like of shape (n_observations, 2)
            An array of bivariate inputs `(u, v)` where each row represents a
            bivariate observation. Both `u` and `v` must be in the interval [0, 1],
            having been transformed to uniform marginals.

        y : None
            Ignored. Provided for compatibility with scikit-learn's API.

        Returns
        -------
        self : IndependentCopula
            Returns the instance itself.
        """
        _ = self._validate_X(X, reset=True)
        return self

    def cdf(self, X: npt.ArrayLike) -> np.ndarray:
        """Compute the CDF of the bivariate Independent copula.

        Parameters
        ----------
        X : array-like of shape (n_observations, 2)
            An array of bivariate inputs `(u, v)` where each row represents a
            bivariate observation. Both `u` and `v` must be in the interval `[0, 1]`,
            having been transformed to uniform marginals.

        Returns
        -------
        cdf : ndarray of shape (n_observations,)
            CDF values for each observation in X.
        """
        skv.check_is_fitted(self)
        X = self._validate_X(X, reset=False)
        cdf = X.prod(axis=1)
        return cdf

    def partial_derivative(
        self, X: npt.ArrayLike, first_margin: bool = False
    ) -> np.ndarray:
        r"""Compute the h-function (partial derivative) for the bivariate Independent
        copula.

        The h-function with respect to the second margin represents the conditional
        distribution function of :math:`u` given :math:`v`:

        .. math::
            \frac{\partial C(u,v)}{\partial v}=u,

        Parameters
        ----------
        X : array-like of shape (n_samples, 2)
            Array of pairs :math:`(u,v)`, where each value is in the interval [0,1].

        Returns
        -------
        np.ndarray
            Array of h-function values for each observation in X.
        """
        skv.check_is_fitted(self)
        X = self._validate_X(X, reset=False)
        h = X[:, 1] if first_margin else X[:, 0]
        return h

    def inverse_partial_derivative(
        self, X: npt.ArrayLike, first_margin: bool = False
    ) -> np.ndarray:
        r"""Compute the inverse of the bivariate copula's partial derivative, commonly
        known as the inverse h-function.

        For the independent copula, the h-function with respect to the second margin is

        .. math::
             h(u\mid v)= u,

        and the derivative with respect to the first margin is

        .. math::
             g(u,v)= v.

        Their inverses are trivial:

          - Given (p,v) for h(u|v)= p, we have u = p.
          - Given (p,u) for g(u,v)= p, we have v = p.

        Parameters
        ----------
        X : array-like of shape (n_observations, 2)
            An array of bivariate inputs `(p, v)`, each in the interval `[0, 1]`.
            - The first column `p` corresponds to the value of the h-function.
            - The second column `v` is the conditioning variable.

        first_margin : bool, default=False
            If True, compute the inverse partial derivative with respect to the first
            margin `u`; otherwise, compute the inverse partial derivative with respect
            to the second margin `v`.

        Returns
        -------
        u : ndarray of shape (n_observations,)
            A 1D-array of length `n_observations`, where each element is the computed
            :math:`u = h^{-1}(p \mid v)` for the corresponding pair in `X`.
        """
        skv.check_is_fitted(self)
        X = self._validate_X(X, reset=False)
        u = X[:, 1] if first_margin else X[:, 0]
        return u

    def score_samples(self, X: npt.ArrayLike) -> np.ndarray:
        """Compute the log-likelihood of each sample (log-pdf) under the model.

        Parameters
        ----------
         X : array-like of shape (n_samples, 2)
            The input data where each row represents a bivariate observation.
            The data should be transformed to uniform marginals in [0, 1].

        Returns
        -------
        density : ndarray of shape (n_samples,)
            The log-likelihood of each sample under the fitted copula.
        """
        skv.check_is_fitted(self)
        X = self._validate_X(X, reset=False)
        return np.zeros(X.shape[0])  # log(1.0)

    @property
    def lower_tail_dependence(self) -> float:
        """Theoretical lower tail dependence coefficient."""
        return 0

    @property
    def upper_tail_dependence(self) -> float:
        """Theoretical upper tail dependence coefficient."""
        return 0

    @property
    def fitted_repr(self) -> str:
        """String representation of the fitted copula."""
        return f"{self.__class__.__name__}()"



"""Expected returns module."""

from skfolio.moments.expected_returns._base import (
    BaseMu,
)
from skfolio.moments.expected_returns._empirical_mu import EmpiricalMu
from skfolio.moments.expected_returns._equilibrium_mu import EquilibriumMu
from skfolio.moments.expected_returns._ew_mu import EWMu
from skfolio.moments.expected_returns._shrunk_mu import ShrunkMu, ShrunkMuMethods

__all__ = [
    "BaseMu",
    "EWMu",
    "EmpiricalMu",
    "EquilibriumMu",
    "ShrunkMu",
    "ShrunkMuMethods",
]



"""Hierarchical Risk Parity Optimization estimator."""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause
# The risk measure generalization and constraint features are derived
# from Riskfolio-Lib, Copyright (c) 2020-2023, Dany Cajas, Licensed under BSD 3 clause.

from __future__ import annotations

import numpy as np
import numpy.typing as npt
import pandas as pd
import scipy.cluster.hierarchy as sch
import sklearn.utils.metadata_routing as skm
import sklearn.utils.validation as skv

import skfolio.typing as skt
from skfolio.cluster import HierarchicalClustering
from skfolio.distance import BaseDistance, PearsonDistance
from skfolio.measures import ExtraRiskMeasure, RiskMeasure
from skfolio.optimization.cluster.hierarchical._base import (
    BaseHierarchicalOptimization,
)
from skfolio.prior import BasePrior, EmpiricalPrior
from skfolio.utils.tools import bisection, check_estimator


class HierarchicalRiskParity(BaseHierarchicalOptimization):
    r"""Hierarchical Risk Parity estimator.

    Hierarchical Risk Parity is a portfolio optimization method developed by Marcos
    Lopez de Prado [1]_.

    This algorithm uses a distance matrix to compute hierarchical clusters using the
    Hierarchical Tree Clustering algorithm. It then employs seriation to rearrange the
    assets in the dendrogram, minimizing the distance between leafs.

    The final step is the recursive bisection where each cluster is split between two
    sub-clusters by starting with the topmost cluster and traversing in a top-down
    manner. For each sub-cluster, we compute the total cluster risk of an inverse-risk
    allocation. A weighting factor is then computed from these two sub-cluster risks,
    which is used to update the cluster weight.

    .. note ::
        The original paper uses the variance as the risk measure and the single-linkage
        method for the Hierarchical Tree Clustering algorithm. Here we generalize it to
        multiple risk measures and linkage methods.
        The default linkage method is set to the Ward
        variance minimization algorithm, which is more stable and has better properties
        than the single-linkage method [2]_.

    Parameters
    ----------
    risk_measure : RiskMeasure or ExtraRiskMeasure, default=RiskMeasure.VARIANCE
        :class:`~skfolio.meta.RiskMeasure` or :class:`~skfolio.meta.ExtraRiskMeasure`
        of the optimization.
        Can be any of:

            * MEAN_ABSOLUTE_DEVIATION
            * FIRST_LOWER_PARTIAL_MOMENT
            * VARIANCE
            * SEMI_VARIANCE
            * CVAR
            * EVAR
            * WORST_REALIZATION
            * CDAR
            * MAX_DRAWDOWN
            * AVERAGE_DRAWDOWN
            * EDAR
            * ULCER_INDEX
            * GINI_MEAN_DIFFERENCE_RATIO
            * VALUE_AT_RISK
            * DRAWDOWN_AT_RISK
            * ENTROPIC_RISK_MEASURE
            * FOURTH_CENTRAL_MOMENT
            * FOURTH_LOWER_PARTIAL_MOMENT

        The default is `RiskMeasure.VARIANCE`.

    prior_estimator : BasePrior, optional
        :ref:`Prior estimator <prior>`.
        The prior estimator is used to estimate the :class:`~skfolio.prior.ReturnDistribution`
        containing the estimation of assets expected returns, covariance matrix and
        returns. The moments and returns estimations are used for the risk computation
        and the returns estimation are used by the distance matrix estimator.
        The default (`None`) is to use :class:`~skfolio.prior.EmpiricalPrior`.

    distance_estimator : BaseDistance, optional
        :ref:`Distance estimator <distance>`.
        The distance estimator is used to estimate the codependence and the distance
        matrix needed for the computation of the linkage matrix.
        The default (`None`) is to use :class:`~skfolio.distance.PearsonDistance`.

    hierarchical_clustering_estimator : HierarchicalClustering, optional
        :ref:`Hierarchical Clustering estimator <hierarchical_clustering>`.
        The hierarchical clustering estimator is used to compute the linkage matrix
        and the hierarchical clustering of the assets based on the distance matrix.
        The default (`None`) is to use
        :class:`~skfolio.cluster.HierarchicalClustering`.

    min_weights : float | dict[str, float] | array-like of shape (n_assets, ), default=0.0
        Minimum assets weights (weights lower bounds). The default is 0.0 (no short
        selling). Negative weights are not allowed. If a float is provided, it is
        applied to each asset. `None` is equivalent to the default `0.0`. If a
        dictionary is provided, its (key/value) pair must be the (asset name/asset
        minimum weight) and the input `X` of the `fit` methods must be a DataFrame with
        the asset names in columns. When using a dictionary, assets values that are not
        provided are assigned the default  minimum weight of `0.0`.

        Example:

           * `min_weights = 0.0` --> long only portfolio (default).
           * `min_weights = {"SX5E": 0.1, "SPX": 0.2}`
           * `min_weights = [0.1, 0.2]`

    max_weights : float | dict[str, float] | array-like of shape (n_assets, ), default=1.0
        Maximum assets weights (weights upper bounds). The default is 1.0 (each asset
        is below 100%). Weights above 1.0 are not allowed. If a float is provided, it is
        applied to each asset. `None` is equivalent to the default `1.0`. If a
        dictionary is provided, its (key/value) pair must be the (asset name/asset
        maximum weight) and the input `X` of the `fit` method must be a DataFrame with
        the asset names in columns. When using a dictionary, assets values that are not
        provided are assigned the default maximum weight of `1.0`.

        Example:

           * `max_weights = 1.0` --> each weight  must be below 100% (default).
           * `max_weights = 0.5` --> each weight must be below 50%.
           * `max_weights = {"SX5E": 0.8, "SPX": 0.9}`
           * `max_weights = [0.8, 0.9]`

    transaction_costs : float | dict[str, float] | array-like of shape (n_assets, ), default=0.0
        Transaction costs of the assets. It is used to add linear transaction costs to
        the optimization problem:

        .. math:: total\_cost = \sum_{i=1}^{N} c_{i} \times |w_{i} - w\_prev_{i}|

        with :math:`c_{i}` the transaction cost of asset i, :math:`w_{i}` its weight
        and :math:`w\_prev_{i}` its previous weight (defined in `previous_weights`).
        The float :math:`total\_cost` is impacting the portfolio expected return in the optimization:

        .. math:: expected\_return = \mu^{T} \cdot w - total\_cost

        with :math:`\mu` the vector af assets' expected returns and :math:`w` the
        vector of assets weights.

        If a float is provided, it is applied to each asset.
        If a dictionary is provided, its (key/value) pair must be the
        (asset name/asset cost) and the input `X` of the `fit` method must be a
        DataFrame with the asset names in columns.
        The default value is `0.0`.

        .. warning::

            Based on the above formula, the periodicity of the transaction costs
            needs to be homogenous to the periodicity of :math:`\mu`. For example, if
            the input `X` is composed of **daily** returns, the `transaction_costs` need
            to be expressed as **daily** costs.
            (See :ref:`sphx_glr_auto_examples_mean_risk_plot_6_transaction_costs.py`)

    management_fees : float | dict[str, float] | array-like of shape (n_assets, ), default=0.0
        Management fees of the assets. It is used to add linear management fees to the
        optimization problem:

        .. math:: total\_fee = \sum_{i=1}^{N} f_{i} \times w_{i}

        with :math:`f_{i}` the management fee of asset i and :math:`w_{i}` its weight.
        The float :math:`total\_fee` is impacting the portfolio expected return in the optimization:

        .. math:: expected\_return = \mu^{T} \cdot w - total\_fee

        with :math:`\mu` the vector af assets expected returns and :math:`w` the vector
        of assets weights.

        If a float is provided, it is applied to each asset.
        If a dictionary is provided, its (key/value) pair must be the
        (asset name/asset fee) and the input `X` of the `fit` method must be a
        DataFrame with the asset names in columns.
        The default value is `0.0`.

        .. warning::

            Based on the above formula, the periodicity of the management fees needs to
            be homogenous to the periodicity of :math:`\mu`. For example, if the input
            `X` is composed of **daily** returns, the `management_fees` need to be
            expressed in **daily** fees.

        .. note::

            Another approach is to directly impact the management fees to the input `X`
            in order to express the returns net of fees. However, when estimating the
            :math:`\mu` parameter using for example Shrinkage estimators, this approach
            would mix a deterministic value with an uncertain one leading to unwanted
            bias in the management fees.

    previous_weights : float | dict[str, float] | array-like of shape (n_assets, ), optional
        Previous weights of the assets. Previous weights are used to compute the
        portfolio total cost. If a float is provided, it is applied to each asset.
        If a dictionary is provided, its (key/value) pair must be the
        (asset name/asset previous weight) and the input `X` of the `fit` method must
        be a DataFrame with the asset names in columns.
        The default (`None`) means no previous weights.
        Additionally, when `fallback="previous_weights"`, failures will fall back to
        these weights if provided.

    portfolio_params : dict, optional
        Portfolio parameters forwarded to the resulting `Portfolio` in `predict`.
        If not provided and if available on the estimator, the following
        attributes are propagated to the portfolio by default: `name`,
        `transaction_costs`, `management_fees`, `previous_weights` and `risk_free_rate`.

    fallback : BaseOptimization | "previous_weights" | list[BaseOptimization | "previous_weights"], optional
        Fallback estimator or a list of estimators to try, in order, when the primary
        optimization raises during `fit`. Alternatively, use `"previous_weights"`
        (alone or in a list) to fall back to the estimator's `previous_weights`.
        When a fallback succeeds, its fitted `weights_` are copied back to the primary
        estimator so that `fit` still returns the original instance. For traceability,
        `fallback_` stores the successful estimator (or the string `"previous_weights"`)
         and `fallback_chain_` stores each attempt with the associated outcome.

    raise_on_failure : bool, default=True
        Controls error handling when fitting fails.
        If True, any failure during `fit` is raised immediately, no `weights_` are
        set and subsequent calls to `predict` will raise a `NotFittedError`.
        If False, errors are not raised; instead, a warning is emitted, `weights_`
        is set to `None` and subsequent calls to `predict` will return a
        `FailedPortfolio`. When fallbacks are specified, this behavior applies only
        after all fallbacks have been exhausted.

    Attributes
    ----------
    weights_ : ndarray of shape (n_assets,)
        Weights of the assets.

    distance_estimator_ : BaseDistance
        Fitted `distance_estimator`.

    hierarchical_clustering_estimator_ : HierarchicalClustering
        Fitted `hierarchical_clustering_estimator`.

    n_features_in_ : int
        Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of assets seen during `fit`. Defined only when `X`
        has asset names that are all strings.

    fallback_ : BaseOptimization | "previous_weights" | None
        The fallback estimator instance, or the string `"previous_weights"`, that
        produced the final result. `None` if no fallback was used.

    fallback_chain_ : list[tuple[str, str]] | None
        Sequence describing the optimization fallback attempts. Each element is a
        pair `(estimator_repr, outcome)` where `estimator_repr` is the string
        representation of the primary estimator or a fallback (e.g. `"EqualWeighted()"`,
        `"previous_weights"`), and `outcome` is `"success"` if that step produced
        a valid solution, otherwise the stringified error message. For successful
        fits without any fallback, this is `None`.

    error_ : str | list[str] | None
        Captured error message(s) when `fit` fails. For multi-portfolio outputs
        (`weights_` is 2D), this is a list aligned with portfolios.

    Notes
    -----
    All estimators should specify all parameters as explicit keyword arguments in
    `__init__` (no `*args` or `**kwargs`), following scikit-learn conventions.

    References
    ----------
    .. [1] "A robust estimator of the efficient frontier",
        SSRN Electronic Journal,
        Marcos López de Prado (2019).

    .. [2] "A review of two decades of correlations, hierarchies, networks and
        clustering in financial markets",
        Gautier Marti, Frank Nielsen, Mikołaj Bińkowski, Philippe Donnat (2020).

    .. [3] "Portfolio Optimization: Theory and Application", Chapter 12,
        Daniel P. Palomar (2025)

    .. [4] "Building diversified portfolios that outperform out of sample",
        The Journal of Portfolio Management,
        Marcos López de Prado (2016).

    .. [5] "Machine Learning for Asset Managers",
        Elements in Quantitative Finance. Cambridge University Press,
        Marcos López de Prado (2020).
    """

    def __init__(
        self,
        risk_measure: RiskMeasure | ExtraRiskMeasure = RiskMeasure.VARIANCE,
        prior_estimator: BasePrior | None = None,
        distance_estimator: BaseDistance | None = None,
        hierarchical_clustering_estimator: HierarchicalClustering | None = None,
        min_weights: skt.MultiInput | None = 0.0,
        max_weights: skt.MultiInput | None = 1.0,
        transaction_costs: skt.MultiInput = 0.0,
        management_fees: skt.MultiInput = 0.0,
        previous_weights: skt.MultiInput | None = None,
        portfolio_params: dict | None = None,
        fallback: skt.Fallback = None,
        raise_on_failure: bool = True,
    ):
        super().__init__(
            risk_measure=risk_measure,
            prior_estimator=prior_estimator,
            distance_estimator=distance_estimator,
            hierarchical_clustering_estimator=hierarchical_clustering_estimator,
            min_weights=min_weights,
            max_weights=max_weights,
            transaction_costs=transaction_costs,
            management_fees=management_fees,
            previous_weights=previous_weights,
            portfolio_params=portfolio_params,
            fallback=fallback,
            raise_on_failure=raise_on_failure,
        )

    def fit(
        self, X: npt.ArrayLike, y: None = None, **fit_params
    ) -> HierarchicalRiskParity:
        """Fit the Hierarchical Risk Parity Optimization estimator.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self : HierarchicalRiskParity
            Fitted estimator.
        """
        routed_params = skm.process_routing(self, "fit", **fit_params)

        # Validate
        if not isinstance(self.risk_measure, RiskMeasure | ExtraRiskMeasure):
            raise TypeError(
                "`risk_measure` must be of type `RiskMeasure` or `ExtraRiskMeasure`"
            )

        if self.risk_measure in [ExtraRiskMeasure.SKEW, ExtraRiskMeasure.KURTOSIS]:
            # Because Skew and Kurtosis can take negative values
            raise ValueError(
                f"risk_measure {self.risk_measure} currently not supported in HRP"
            )

        self.prior_estimator_ = check_estimator(
            self.prior_estimator,
            default=EmpiricalPrior(),
            check_type=BasePrior,
        )
        self.distance_estimator_ = check_estimator(
            self.distance_estimator,
            default=PearsonDistance(),
            check_type=BaseDistance,
        )
        self.hierarchical_clustering_estimator_ = check_estimator(
            self.hierarchical_clustering_estimator,
            default=HierarchicalClustering(),
            check_type=HierarchicalClustering,
        )

        # Fit the estimators
        self.prior_estimator_.fit(X, y, **routed_params.prior_estimator.fit)
        return_distribution = self.prior_estimator_.return_distribution_
        returns = return_distribution.returns

        # To keep the asset_names
        if isinstance(X, pd.DataFrame):
            returns = pd.DataFrame(returns, columns=X.columns)

        # noinspection PyArgumentList
        self.distance_estimator_.fit(returns, y, **routed_params.distance_estimator.fit)
        distance = self.distance_estimator_.distance_

        # To keep the asset_names
        if isinstance(X, pd.DataFrame):
            distance = pd.DataFrame(distance, columns=X.columns)

        # noinspection PyArgumentList
        self.hierarchical_clustering_estimator_.fit(
            X=distance, y=None, **routed_params.hierarchical_clustering_estimator.fit
        )

        X = skv.validate_data(self, X)
        n_assets = X.shape[1]

        min_weights, max_weights = self._convert_weights_bounds(n_assets=n_assets)
        assets_risks = self._unitary_risks(return_distribution=return_distribution)

        ordered_linkage_matrix = sch.optimal_leaf_ordering(
            self.hierarchical_clustering_estimator_.linkage_matrix_,
            self.hierarchical_clustering_estimator_.condensed_distance_,
        )
        sorted_assets = sch.leaves_list(ordered_linkage_matrix)

        weights = np.ones(n_assets)
        items = [sorted_assets]

        while len(items) > 0:
            new_items = []
            for clusters_ids in bisection(items):
                new_items += clusters_ids
                risks = []
                for ids in clusters_ids:
                    inv_risk_w = np.zeros(n_assets)
                    inv_risk_w[ids] = 1 / assets_risks[ids]
                    inv_risk_w /= inv_risk_w.sum()
                    risks.append(
                        self._risk(
                            weights=inv_risk_w, return_distribution=return_distribution
                        )
                    )
                left_risk, right_risk = risks
                left_cluster, right_cluster = clusters_ids
                alpha = 1 - left_risk / (left_risk + right_risk)
                # Weights constraints
                alpha = _apply_weight_constraints_to_split_factor(
                    alpha=alpha,
                    weights=weights,
                    max_weights=max_weights,
                    min_weights=min_weights,
                    left_cluster=left_cluster,
                    right_cluster=right_cluster,
                )
                weights[left_cluster] *= alpha
                weights[right_cluster] *= 1 - alpha
            items = new_items

        self.weights_ = weights
        return self


def _apply_weight_constraints_to_split_factor(
    alpha: float,
    max_weights: np.ndarray,
    min_weights: np.ndarray,
    weights: np.ndarray,
    left_cluster: np.ndarray,
    right_cluster: np.ndarray,
) -> float:
    """
    Apply weight constraints to the split factor alpha of the ,Hierarchical Tree
    Clustering algorithm.

    Parameters
    ----------
    alpha : float
        The split factor alpha of the Hierarchical Tree Clustering algorithm.

    min_weights : ndarray of shape (n_assets,)
        The weight lower bound 1D array.

    max_weights : ndarray of shape (n_assets,)
        The weight upper bound 1D array.

    weights : np.ndarray of shape (n_assets,)
        The assets weights.

    left_cluster : ndarray of shape (n_left_cluster,)
        Indices of the left cluster weights.

    right_cluster : ndarray of shape (n_right_cluster,)
        Indices of the right cluster weights.

    Returns
    -------
    value : float
        The transformed split factor alpha incorporating the weight constraints.
    """
    alpha = min(
        np.sum(max_weights[left_cluster]) / weights[left_cluster[0]],
        max(np.sum(min_weights[left_cluster]) / weights[left_cluster[0]], alpha),
    )
    alpha = 1 - min(
        np.sum(max_weights[right_cluster]) / weights[right_cluster[0]],
        max(
            np.sum(min_weights[right_cluster]) / weights[right_cluster[0]],
            1 - alpha,
        ),
    )
    return alpha



"""Shrunk Covariance Estimators."""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause
# Implementation derived from:
# scikit-learn, Copyright (c) 2007-2010 David Cournapeau, Fabian Pedregosa, Olivier
# Grisel Licensed under BSD 3 clause.

import numpy.typing as npt
import sklearn.covariance as skc

from skfolio.moments.covariance._base import BaseCovariance


class ShrunkCovariance(BaseCovariance, skc.ShrunkCovariance):
    """Covariance estimator with shrinkage.

    Read more in `scikit-learn
    <https://scikit-learn.org/stable/modules/generated/sklearn.covariance.ShrunkCovariance.html>`_.

    Parameters
    ----------
    store_precision : bool, default=True
        Specify if the estimated precision is stored.

    assume_centered : bool, default=False
        If True, data will not be centered before computation.
        Useful when working with data whose mean is almost, but not exactly
        zero.
        If False (default), data will be centered before computation.

    shrinkage : float, default=0.1
        Coefficient in the convex combination used for the computation
        of the shrunk estimate. Range is [0, 1].

    Attributes
    ----------
    covariance_ : ndarray of shape (n_assets, n_assets)
        Estimated covariance.

    location_ : ndarray of shape (n_assets,)
        Estimated location, i.e. the estimated mean.

    precision_ : ndarray of shape (n_assets, n_assets)
        Estimated pseudo inverse matrix.
        (stored only if store_precision is True)

    n_features_in_ : int
       Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
       Names of features seen during `fit`. Defined only when `X`
       has feature names that are all strings.

    Notes
    -----
    The regularized covariance is given by:

    (1 - shrinkage) * cov + shrinkage * mu * np.identity(n_features)

    where mu = trace(cov) / n_features
    """

    def __init__(
        self,
        store_precision=True,
        assume_centered=False,
        shrinkage=0.1,
        nearest: bool = True,
        higham: bool = False,
        higham_max_iteration: int = 100,
    ):
        super().__init__(
            nearest=nearest,
            higham=higham,
            higham_max_iteration=higham_max_iteration,
        )
        skc.ShrunkCovariance.__init__(
            self,
            store_precision=store_precision,
            assume_centered=assume_centered,
            shrinkage=shrinkage,
        )

    def fit(self, X: npt.ArrayLike, y=None) -> "ShrunkCovariance":
        """Fit the shrunk covariance model to X.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
         Price returns of the assets.

        y : Ignored
          Not used, present for API consistency by convention.

        Returns
        -------
        self : ShrunkCovariance
          Fitted estimator.
        """
        skc.ShrunkCovariance.fit(self, X)
        self._set_covariance(self.covariance_)
        return self



"""Opinion Pooling estimator."""

# Copyright (c) 2025
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# Credits: Vincent Maladière, Matteo Manzi, Carlo Nicolini
# SPDX-License-Identifier: BSD-3-Clause

import numpy as np
import numpy.typing as npt
import pandas as pd
import scipy.special as scs
import scipy.special as sp
import sklearn as sk
import sklearn.utils as sku
import sklearn.utils.metadata_routing as skm
import sklearn.utils.parallel as skp
import sklearn.utils.validation as skv

import skfolio.measures as sm
from skfolio.prior._base import BasePrior, ReturnDistribution
from skfolio.utils.composition import BaseComposition
from skfolio.utils.tools import check_estimator, fit_single_estimator


class OpinionPooling(BasePrior, BaseComposition):
    r"""Opinion Pooling estimator.

    Opinion Pooling (also called Belief Aggregation or Risk Aggregation) is a process
    in which different probability distributions (opinions), produced by different
    experts, are combined to yield a single probability distribution (consensus).

    Expert opinions (also called individual prior distributions) can be
    **elicited** from domain experts or **derived** from quantitative analyses.

    The `OpinionPooling` estimator takes a list of prior estimators, each of which
    produces scenario probabilities (which we use as `sample_weight`), and pools them
    into a single consensus probability .

    You can choose between linear (arithmetic) pooling or logarithmic (geometric)
    pooling, and optionally apply robust pooling using a Kullback-Leibler divergence
    penalty to down-weight experts whose views deviate strongly from the group
    consensus.

    Parameters
    ----------
    estimators : list of (str, BasePrior)
        A list of :ref:`prior estimators <prior>` representing opinions to be pooled
        into a single consensus.
        Each element of the list is defined as a tuple of string (i.e. name) and an
        estimator instance. Each must expose `sample_weight` such as in
        :class:`~skfolio.prior.EntropyPooling`.

    opinion_probabilities : array-like of float, optional
        Probability mass assigned to each opinion, in [0,1] summing to ≤1.
        Any leftover mass is assigned to the uniform (uninformative) prior.
        The default (None), is to assign the same probability to each opinion.

    prior_estimator : BasePrior, optional
        Common prior for all `estimators`. If provided, each estimator from `estimators`
        will be fitted using this common prior before pooling. Setting `prior_estimator`
        inside individual `estimators` is disabled to avoid mixing different prior
        scenarios (each estimator must have the same underlying distribution).
        For example, using `prior_estimator = SyntheticData(n_samples=10_000)` will
        generate 10,000 synthetic data points from a Vine Copula before fitting the
        estimators on this common distribution.

    is_linear_pooling : bool, default=True
        If True, combine each opinion via Linear Opinion Pooling
        (arithmetic mean); if False, use Logarithmic Opinion Pooling (geometric
        mean).

        Linear Opinion Pooling:
            * Retains all nonzero support (no "zero-forcing").
            * Produces an averaging that is more evenly spread across all expert opinions.

        Logarithmic Opinion Pooling:
            * Zero-Preservation. Any scenario assigned zero probability by any expert
              remains zero in the aggregate.
            * Information-Theoretic Optimality. Yields the distribution that minimizes
              the weighted sum of KL-divergences from each expert's distribution.
            * Robust to Extremes: down-weight extreme or contrarian views more severely.

    divergence_penalty : float, default=0.0
        Non-negative factor (:math:`\alpha`) that penalizes each opinion's divergence
        from the group consensus, yielding more robust pooling.
        A higher value more strongly down-weights deviating opinions.

        The robust opinion probabilities are given by:

        .. math::
            \tilde{p}_i = \frac{p_i \exp\bigl(-\alpha D_i\bigr)}
            {\displaystyle \sum_{k=1}^N p_k \exp\bigl(-\alpha D_k\bigr)}
            \quad\text{for }i = 1,\dots,N

        where

        * :math:`N` is the number of experts `len(estimators)`

        * :math:`M` is the number of scenarios `len(observations)`

        * :math:`D_i` is the KL-divergence of expert *i*'s distribution from consensus:

          .. math::
             D_i = \mathrm{KL}\bigl(w_i \,\|\, c\bigr)
                 = \sum_{j=1}^M w_{ij}\,\ln\!\frac{w_{ij}}{c_j}
             \quad\text{for }i = 1,\dots,N.

        * :math:`w_i` is the sample-weight vector (scenario probabilities) from expert
          *i*,  with :math:`\sum_{j=1}^M w_{ij} = 1`.

        * :math:`p_i` is the initial opinion probability of expert *i*, with
          :math:`\sum_{i=1}^N p_i \le 1` (any leftover mass goes to a uniform prior).

        * :math:`c_j` is the consensus of scenario :math:`j`:

          .. math::
             c_j = \sum_{i=1}^N p_i \, w_{ij} \quad\text{for }j = 1,\dots,M.

    n_jobs : int, optional
        The number of jobs to run in parallel for `fit` of all `estimators`.
        The value `-1` means using all processors.
        The default (`None`) means 1 unless in a `joblib.parallel_backend` context.

    Attributes
    ----------
    return_distribution_ : ReturnDistribution
        Fitted :class:`~skfolio.prior.ReturnDistribution` to be used by the optimization
        estimators, containing the assets distribution, moments estimation and the
        opinion-pooling sample weights.

    estimators_ : list[BasePrior]
        The elements of the `estimators` parameter, having been fitted on the
        training data.

    named_estimators_ : dict[str, BasePrior]
        Attribute to access any fitted sub-estimators by name.

    prior_estimator_ : BasePrior
        Fitted `prior_estimator` if provided.

    opinion_probabilities_ : ndarray of shape (n_opinions,)
        Final opinion probabilities after applying the KL-divergence penalty.
        If the initial `opinion_probabilities` doesn't sum to one, the last element of
        `opinion_probabilities_` is the probability assigned to the uniform prior.

    n_features_in_ : int
        Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of assets seen during `fit`. Defined only when `X`
        has assets names that are all strings.

    References
    ----------
    .. [1] "Probabilistic opinion pooling generalized",
            Social Choice and Welfare, Dietrich & List (2017)

    .. [2] "Opinion Aggregation and Individual Expertise",
            Oxford University Press, Martini & Sprenger (2017)

    .. [3] "Rational Decisions",
            Journal of the Royal Statistical Society, Good  (1952)

    Examples
    --------
    For a full tutorial on entropy pooling, see :ref:`sphx_glr_auto_examples_entropy_pooling_plot_2_opinion_pooling.py`.

    >>> from skfolio import RiskMeasure
    >>> from skfolio.datasets import load_sp500_dataset
    >>> from skfolio.preprocessing import prices_to_returns
    >>> from skfolio.prior import EntropyPooling, OpinionPooling
    >>> from skfolio.optimization import RiskBudgeting
    >>>
    >>> prices = load_sp500_dataset()
    >>> X = prices_to_returns(prices)
    >>>
    >>> # We consider two expert opinions, each generated via Entropy Pooling with
    >>> # user-defined views.
    >>> # We assign probabilities of 40% to Expert 1, 50% to Expert 2, and by default
    >>> # the remaining 10% is allocated to the prior distribution:
    >>> opinion_1 = EntropyPooling(cvar_views=["AMD == 0.10"])
    >>> opinion_2 = EntropyPooling(
    ...     mean_views=["AMD >= BAC", "JPM <= prior(JPM) * 0.8"],
    ...     cvar_views=["GE == 0.12"],
    ... )
    >>>
    >>> opinion_pooling = OpinionPooling(
    ...     estimators=[("opinion_1", opinion_1), ("opinion_2", opinion_2)],
    ...     opinion_probabilities=[0.4, 0.5],
    ... )
    >>>
    >>> opinion_pooling.fit(X)
    >>>
    >>> print(opinion_pooling.return_distribution_.sample_weight)
    >>>
    >>> # CVaR Risk Parity optimization on opinion Pooling
    >>> model = RiskBudgeting(
    ...     risk_measure=RiskMeasure.CVAR,
    ...     prior_estimator=opinion_pooling
    ... )
    >>> model.fit(X)
    >>> print(model.weights_)
    >>>
    >>> # Stress Test the Portfolio
    >>> opinion_1 = EntropyPooling(cvar_views=["AMD == 0.05"])
    >>> opinion_2 = EntropyPooling(cvar_views=["AMD == 0.10"])
    >>> opinion_pooling = OpinionPooling(
    ...     estimators=[("opinion_1", opinion_1), ("opinion_2", opinion_2)],
    ...     opinion_probabilities=[0.6, 0.4],
    ... )
    >>> opinion_pooling.fit(X)
    >>>
    >>> stressed_dist = opinion_pooling.return_distribution_
    >>>
    >>> stressed_ptf = model.predict(stressed_dist)
    """

    estimators_: list[BasePrior]
    named_estimators_: dict[str, BasePrior]
    opinion_probabilities_: np.ndarray
    prior_estimator_: BasePrior
    n_features_in_: int
    feature_names_in_: np.ndarray

    def __init__(
        self,
        estimators: list[tuple[str, BasePrior]],
        opinion_probabilities: list[float] | None = None,
        prior_estimator: BasePrior | None = None,
        is_linear_pooling: bool = True,
        divergence_penalty: float = 0.0,
        n_jobs: int | None = None,
    ):
        self.estimators = estimators
        self.opinion_probabilities = opinion_probabilities
        self.prior_estimator = prior_estimator
        self.divergence_penalty = divergence_penalty
        self.is_linear_pooling = is_linear_pooling
        self.n_jobs = n_jobs

    @property
    def named_estimators(self):
        """Dictionary to access any fitted sub-estimators by name.

        Returns
        -------
        :class:`~sklearn.utils.Bunch`
        """
        return sku.Bunch(**dict(self.estimators))

    def _validate_estimators(self) -> tuple[list[str], list[BasePrior]]:
        """Validate the `estimators` parameter.

        Returns
        -------
        names : list[str]
            The list of estimators names.
        estimators : list[BaseOptimization
            The list of optimization estimators.
        """
        if self.estimators is None or len(self.estimators) == 0:
            raise ValueError(
                "Invalid 'estimators' attribute, 'estimators' should be a list"
                " of (string, estimator) tuples."
            )
        names, estimators = zip(*self.estimators, strict=True)
        # defined by MetaEstimatorMixin
        self._validate_names(names)

        for estimator in estimators:
            if getattr(estimator, "prior_estimator", None) is not None:
                raise ValueError(
                    "Cannot set `prior_estimator` on individual estimators within "
                    "`OpinionPooling` to avoid mixing different prior scenarios. "
                    "Please leave those as `None` and specify your prior directly via "
                    "the `prior_estimator` parameter of the `OpinionPooling` class."
                )

        return names, estimators

    def set_params(self, **params):
        """Set the parameters of an estimator from the ensemble.

        Valid parameter keys can be listed with `get_params()`. Note that you
        can directly set the parameters of the estimators contained in
        `estimators`.

        Parameters
        ----------
        **params : keyword arguments
            Specific parameters using e.g.
            `set_params(parameter_name=new_value)`. In addition, to setting the
            parameters of the estimator, the individual estimator of the
            estimators can also be set, or can be removed by setting them to
            'drop'.

        Returns
        -------
        self : object
            Estimator instance.
        """
        super()._set_params("estimators", **params)
        return self

    def get_params(self, deep=True):
        """Get the parameters of an estimator from the ensemble.

        Returns the parameters given in the constructor as well as the
        estimators contained within the `estimators` parameter.

        Parameters
        ----------
        deep : bool, default=True
            Setting it to True gets the various estimators and the parameters
            of the estimators as well.

        Returns
        -------
        params : dict
            Parameter and estimator names mapped to their values or parameter
            names mapped to their values.
        """
        return super()._get_params("estimators", deep=deep)

    def get_metadata_routing(self):
        router = skm.MetadataRouter(owner=self.__class__.__name__)
        for name, estimator in self.estimators:
            router.add(
                **{name: estimator},
                method_mapping=skm.MethodMapping().add(caller="fit", callee="fit"),
            )
        return router

    def fit(self, X: npt.ArrayLike, y=None, **fit_params) -> "OpinionPooling":
        """Fit the Opinion Pooling estimator.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
           Price returns of the assets.

        y : Ignored
           Not used, present for API consistency by convention.

        **fit_params : dict
            Parameters to pass to the underlying estimators.
            Only available if `enable_metadata_routing=True`, which can be
            set by using ``sklearn.set_config(enable_metadata_routing=True)``.
            See :ref:`Metadata Routing User Guide <metadata_routing>` for
            more details.

        Returns
        -------
        self : OpinionPooling
           Fitted estimator.
        """
        routed_params = skm.process_routing(self, "fit", **fit_params)

        skv.validate_data(self, X)

        names, all_estimators = self._validate_estimators()

        opinion_probabilities = self._validate_opinion_probabilities()

        if self.prior_estimator is not None:
            self.prior_estimator_ = check_estimator(
                self.prior_estimator,
                default=None,
                check_type=BasePrior,
            )
            # fitting prior estimator
            self.prior_estimator_.fit(X, y, **routed_params.prior_estimator.fit)
            returns = self.prior_estimator_.return_distribution_.returns
            # To keep the asset_names
            if hasattr(self, "feature_names_in_"):
                returns = pd.DataFrame(returns, columns=self.feature_names_in_)
        else:
            returns = X

        # Fit the prior estimators on the whole training data. Those
        # prior estimators will be used to retrieve the sample weights.
        self.estimators_ = skp.Parallel(n_jobs=self.n_jobs)(
            skp.delayed(fit_single_estimator)(
                sk.clone(est), returns, None, routed_params[name]["fit"]
            )
            for name, est in zip(names, all_estimators, strict=True)
        )

        self.named_estimators_ = {
            name: estimator
            for name, estimator in zip(names, self.estimators_, strict=True)
        }

        sample_weights = []
        for estimator in self.estimators_:
            if estimator.return_distribution_.sample_weight is None:
                raise ValueError(
                    f"Estimator `{estimator.__class__.__name__}` did not produce "
                    "a `return_distribution_.sample_weight`. OpinionPooling requires "
                    "each estimator to expose sample weights (e.g. via EntropyPooling)."
                )
            sample_weights.append(estimator.return_distribution_.sample_weight)
        sample_weights = np.array(sample_weights)

        returns = np.asarray(returns)
        n_observations = len(returns)

        # Add the remaining part of the opinion_probabilities to the uniform prior
        q_weight = 1.0 - opinion_probabilities.sum()
        if q_weight > 1e-8:
            opinion_probabilities = np.append(opinion_probabilities, q_weight)
            q = np.ones(n_observations) / n_observations
            sample_weights = np.vstack((sample_weights, q))

        opinion_probabilities = self._compute_robust_opinion_probabilities(
            opinion_probabilities=opinion_probabilities, sample_weights=sample_weights
        )

        if self.is_linear_pooling:
            sample_weight = opinion_probabilities @ sample_weights
        else:
            # let exact 0 in sample weights flow through
            with np.errstate(divide="ignore"):
                u = opinion_probabilities @ np.log(sample_weights)
                sample_weight = np.exp(u - sp.logsumexp(u))

        self.opinion_probabilities_ = opinion_probabilities
        self.return_distribution_ = ReturnDistribution(
            mu=sm.mean(returns, sample_weight=sample_weight),
            covariance=np.cov(returns, rowvar=False, aweights=sample_weight),
            returns=returns,
            sample_weight=sample_weight,
        )
        return self

    def _validate_opinion_probabilities(self) -> np.ndarray:
        """Validate `opinion_probabilities`."""
        n_opinions = len(self.estimators)
        if self.opinion_probabilities is None:
            return np.ones(n_opinions) / n_opinions

        opinion_probabilities = np.asarray(self.opinion_probabilities)

        if len(opinion_probabilities) != n_opinions:
            raise ValueError(
                f"`opinion_probabilities` length ({len(opinion_probabilities)}) "
                f"does not match number of estimators ({n_opinions})."
            )

        if np.any(opinion_probabilities < 0) or np.any(opinion_probabilities > 1):
            raise ValueError(
                "`The entries of `opinion_probabilities` must be between 0 and 1"
            )
        if opinion_probabilities.sum() > 1.0:
            raise ValueError(
                "The entries of `opinion_probabilities` must sum to at most 1; "
                "any remaining mass (1-sum) is allocated to the uniform prior."
            )
        return opinion_probabilities

    def _compute_robust_opinion_probabilities(
        self, opinion_probabilities: np.ndarray, sample_weights: np.ndarray
    ) -> np.ndarray:
        """Compute the robust `opinion_probabilities` using KL-divergence."""
        if self.divergence_penalty < 0:
            raise ValueError("`divergence_penalty` cannot be negative")

        if self.divergence_penalty == 0:
            return opinion_probabilities

        consensus = opinion_probabilities @ sample_weights
        divergences = np.sum(scs.rel_entr(sample_weights, consensus), axis=1)
        opinion_probabilities *= np.exp(-self.divergence_penalty * divergences)
        opinion_probabilities /= opinion_probabilities.sum()
        return opinion_probabilities
